@@ -409,6 +409,15 @@ CanonicalFilePath ReadTransaction::Implementation::getFullFilePathFromCode(IDCod
   return CanonicalFilePathRef::getAsCanonicalPath(path);
 }
 
+std::string ReadTransaction::Implementation::getUnitFileIdentifierFromCode(IDCode filePathCode) {
+  std::string path;
+  {
+    llvm::raw_string_ostream OS(path);
+    getFullFilePathFromCode(filePathCode, OS);
+  }
+  return path;
+}
+
 CanonicalFilePathRef ReadTransaction::Implementation::getDirectoryFromCode(IDCode dirCode) {
   lmdb::val key{&dirCode, sizeof(dirCode)};
   lmdb::val value{};
@@ -486,6 +495,10 @@ IDCode ReadTransaction::Implementation::getFilePathCode(CanonicalFilePathRef fil
   return makeIDCodeFromString(filePath.getPath());
 }
 
+IDCode ReadTransaction::Implementation::getUnitPathCode(StringRef filePath) {
+  return makeIDCodeFromString(filePath);
+}
+
 bool ReadTransaction::Implementation::getFilePathFromValue(lmdb::val &filePathValue, raw_ostream &OS) {
   auto &db = DBase->impl();
   auto &dbiDirNames = db.getDBIDirNameByCode();
@@ -721,6 +734,10 @@ CanonicalFilePath ReadTransaction::getFullFilePathFromCode(IDCode filePathCode)
   return Impl->getFullFilePathFromCode(filePathCode);
 }
 
+std::string ReadTransaction::getUnitFileIdentifierFromCode(IDCode filePathCode) {
+  return Impl->getUnitFileIdentifierFromCode(filePathCode);
+}
+
 CanonicalFilePathRef ReadTransaction::getDirectoryFromCode(IDCode dirCode) {
   return Impl->getDirectoryFromCode(dirCode);
 }
@@ -738,6 +755,10 @@ IDCode ReadTransaction::getFilePathCode(CanonicalFilePathRef filePath) {
   return Impl->getFilePathCode(filePath);
 }
 
+IDCode ReadTransaction::getUnitFileIdentifierCode(StringRef filePath) {
+  return Impl->getUnitPathCode(filePath);
+}
+
 UnitInfo ReadTransaction::getUnitInfo(IDCode unitCode) {
   return Impl->getUnitInfo(unitCode);
 }
