@@ -6,6 +6,9 @@
 //
 //===----------------------------------------------------------------------===//
 
+#ifndef LLVM_LIBC_FUZZING_MATH_REMQUO_H
+#define LLVM_LIBC_FUZZING_MATH_REMQUO_H
+
 #include "src/__support/FPUtil/FPBits.h"
 
 #include <math.h>
@@ -46,3 +49,5 @@ void RemQuoDiff(RemQuoFunc<T> func1, RemQuoFunc<T> func2, const uint8_t *data,
   if (bits1.uintval() != bits2.uintval())
     __builtin_trap();
 }
+
+#endif // LLVM_LIBC_FUZZING_MATH_REMQUO_H
