@@ -11,13 +11,15 @@
 //
 //===----------------------------------------------------------------------===//
 
-#include "PassDetail.h"
-#include "mlir/Conversion/SCFToStandard/SCFToStandard.h"
-#include "mlir/Dialect/Arithmetic/IR/Arithmetic.h"
-#include "mlir/Dialect/Async/IR/Async.h"
 #include "mlir/Dialect/Async/Passes.h"
-#include "mlir/Dialect/SCF/SCF.h"
-#include "mlir/Dialect/StandardOps/IR/Ops.h"
+
+#include "PassDetail.h"
+#include "mlir/Conversion/SCFToControlFlow/SCFToControlFlow.h"
+#include "mlir/Dialect/Arith/IR/Arith.h"
+#include "mlir/Dialect/Async/IR/Async.h"
+#include "mlir/Dialect/ControlFlow/IR/ControlFlowOps.h"
+#include "mlir/Dialect/Func/IR/FuncOps.h"
+#include "mlir/Dialect/SCF/IR/SCF.h"
 #include "mlir/IR/BlockAndValueMapping.h"
 #include "mlir/IR/ImplicitLocOpBuilder.h"
 #include "mlir/IR/PatternMatch.h"
@@ -26,6 +28,11 @@
 #include "llvm/ADT/SetVector.h"
 #include "llvm/Support/Debug.h"
 
+namespace mlir {
+#define GEN_PASS_DEF_ASYNCTOASYNCRUNTIME
+#include "mlir/Dialect/Async/Passes.h.inc"
+} // namespace mlir
+
 using namespace mlir;
 using namespace mlir::async;
 
@@ -36,7 +43,7 @@ static constexpr const char kAsyncFnPrefix[] = "async_execute_fn";
 namespace {
 
 class AsyncToAsyncRuntimePass
-    : public AsyncToAsyncRuntimeBase<AsyncToAsyncRuntimePass> {
+    : public impl::AsyncToAsyncRuntimeBase<AsyncToAsyncRuntimePass> {
 public:
   AsyncToAsyncRuntimePass() = default;
   void runOnOperation() override;
@@ -55,7 +62,7 @@ public:
 /// operation to enable non-blocking waiting via coroutine suspension.
 namespace {
 struct CoroMachinery {
-  FuncOp func;
+  func::FuncOp func;
 
   // Async execute region returns a completion token, and an async value for
   // each yielded value.
@@ -67,7 +74,7 @@ struct CoroMachinery {
   Value asyncToken; // token representing completion of the async region
   llvm::SmallVector<Value, 4> returnValues; // returned async values
 
-  Value coroHandle; // coroutine handle (!async.coro.handle value)
+  Value coroHandle; // coroutine handle (!async.coro.getHandle value)
   Block *entry;     // coroutine entry block
   Block *setError;  // switch completion token and all values to error state
   Block *cleanup;   // coroutine cleanup block
@@ -103,27 +110,27 @@ struct CoroMachinery {
 ///     ^entry(<function-arguments>):
 ///       %token = <async token> : !async.token    // create async runtime token
 ///       %value = <async value> : !async.value<T> // create async value
-///       %id = async.coro.id                      // create a coroutine id
+///       %id = async.coro.getId                      // create a coroutine id
 ///       %hdl = async.coro.begin %id              // create a coroutine handle
-///       br ^preexisting_entry_block
+///       cf.br ^preexisting_entry_block
 ///
 ///     /*  preexisting blocks modified to branch to the cleanup block */
 ///
 ///     ^set_error: // this block created lazily only if needed (see code below)
 ///       async.runtime.set_error %token : !async.token
 ///       async.runtime.set_error %value : !async.value<T>
-///       br ^cleanup
+///       cf.br ^cleanup
 ///
 ///     ^cleanup:
 ///       async.coro.free %hdl // delete the coroutine state
-///       br ^suspend
+///       cf.br ^suspend
 ///
 ///     ^suspend:
 ///       async.coro.end %hdl // marks the end of a coroutine
 ///       return %token, %value : !async.token, !async.value<T>
 ///   }
 ///
-static CoroMachinery setupCoroMachinery(FuncOp func) {
+static CoroMachinery setupCoroMachinery(func::FuncOp func) {
   assert(!func.getBlocks().empty() && "Function must have an entry block");
 
   MLIRContext *ctx = func.getContext();
@@ -135,19 +142,21 @@ static CoroMachinery setupCoroMachinery(FuncOp func) {
   // ------------------------------------------------------------------------ //
   // Allocate async token/values that we will return from a ramp function.
   // ------------------------------------------------------------------------ //
-  auto retToken = builder.create<RuntimeCreateOp>(TokenType::get(ctx)).result();
+  auto retToken =
+      builder.create<RuntimeCreateOp>(TokenType::get(ctx)).getResult();
 
   llvm::SmallVector<Value, 4> retValues;
   for (auto resType : func.getCallableResults().drop_front())
-    retValues.emplace_back(builder.create<RuntimeCreateOp>(resType).result());
+    retValues.emplace_back(
+        builder.create<RuntimeCreateOp>(resType).getResult());
 
   // ------------------------------------------------------------------------ //
   // Initialize coroutine: get coroutine id and coroutine handle.
   // ------------------------------------------------------------------------ //
   auto coroIdOp = builder.create<CoroIdOp>(CoroIdType::get(ctx));
   auto coroHdlOp =
-      builder.create<CoroBeginOp>(CoroHandleType::get(ctx), coroIdOp.id());
-  builder.create<BranchOp>(originalEntryBlock);
+      builder.create<CoroBeginOp>(CoroHandleType::get(ctx), coroIdOp.getId());
+  builder.create<cf::BranchOp>(originalEntryBlock);
 
   Block *cleanupBlock = func.addBlock();
   Block *suspendBlock = func.addBlock();
@@ -156,10 +165,10 @@ static CoroMachinery setupCoroMachinery(FuncOp func) {
   // Coroutine cleanup block: deallocate coroutine frame, free the memory.
   // ------------------------------------------------------------------------ //
   builder.setInsertionPointToStart(cleanupBlock);
-  builder.create<CoroFreeOp>(coroIdOp.id(), coroHdlOp.handle());
+  builder.create<CoroFreeOp>(coroIdOp.getId(), coroHdlOp.getHandle());
 
   // Branch into the suspend block.
-  builder.create<BranchOp>(suspendBlock);
+  builder.create<cf::BranchOp>(suspendBlock);
 
   // ------------------------------------------------------------------------ //
   // Coroutine suspend block: mark the end of a coroutine and return allocated
@@ -168,33 +177,38 @@ static CoroMachinery setupCoroMachinery(FuncOp func) {
   builder.setInsertionPointToStart(suspendBlock);
 
   // Mark the end of a coroutine: async.coro.end
-  builder.create<CoroEndOp>(coroHdlOp.handle());
+  builder.create<CoroEndOp>(coroHdlOp.getHandle());
 
   // Return created `async.token` and `async.values` from the suspend block.
   // This will be the return value of a coroutine ramp function.
   SmallVector<Value, 4> ret{retToken};
   ret.insert(ret.end(), retValues.begin(), retValues.end());
-  builder.create<ReturnOp>(ret);
+  builder.create<func::ReturnOp>(ret);
 
   // `async.await` op lowering will create resume blocks for async
   // continuations, and will conditionally branch to cleanup or suspend blocks.
 
-  for (Block &block : func.body().getBlocks()) {
+  for (Block &block : func.getBody().getBlocks()) {
     if (&block == entryBlock || &block == cleanupBlock ||
         &block == suspendBlock)
       continue;
     Operation *terminator = block.getTerminator();
     if (auto yield = dyn_cast<YieldOp>(terminator)) {
       builder.setInsertionPointToEnd(&block);
-      builder.create<BranchOp>(cleanupBlock);
+      builder.create<cf::BranchOp>(cleanupBlock);
     }
   }
 
+  // The switch-resumed API based coroutine should be marked with
+  // coroutine.presplit attribute to mark the function as a coroutine.
+  func->setAttr("passthrough", builder.getArrayAttr(
+                                   StringAttr::get(ctx, "presplitcoroutine")));
+
   CoroMachinery machinery;
   machinery.func = func;
   machinery.asyncToken = retToken;
   machinery.returnValues = retValues;
-  machinery.coroHandle = coroHdlOp.handle();
+  machinery.coroHandle = coroHdlOp.getHandle();
   machinery.entry = entryBlock;
   machinery.setError = nullptr; // created lazily only if needed
   machinery.cleanup = cleanupBlock;
@@ -220,7 +234,7 @@ static Block *setupSetErrorBlock(CoroMachinery &coro) {
     builder.create<RuntimeSetErrorOp>(retValue);
 
   // Branch into the cleanup block.
-  builder.create<BranchOp>(coro.cleanup);
+  builder.create<cf::BranchOp>(coro.cleanup);
 
   return coro.setError;
 }
@@ -229,7 +243,7 @@ static Block *setupSetErrorBlock(CoroMachinery &coro) {
 /// function.
 ///
 /// Note that this is not reversible transformation.
-static std::pair<FuncOp, CoroMachinery>
+static std::pair<func::FuncOp, CoroMachinery>
 outlineExecuteOp(SymbolTable &symbolTable, ExecuteOp execute) {
   ModuleOp module = execute->getParentOfType<ModuleOp>();
 
@@ -238,13 +252,14 @@ outlineExecuteOp(SymbolTable &symbolTable, ExecuteOp execute) {
 
   // Make sure that all constants will be inside the outlined async function to
   // reduce the number of function arguments.
-  cloneConstantsIntoTheRegion(execute.body());
+  cloneConstantsIntoTheRegion(execute.getBodyRegion());
 
   // Collect all outlined function inputs.
-  SetVector<mlir::Value> functionInputs(execute.dependencies().begin(),
-                                        execute.dependencies().end());
-  functionInputs.insert(execute.operands().begin(), execute.operands().end());
-  getUsedValuesDefinedAbove(execute.body(), functionInputs);
+  SetVector<mlir::Value> functionInputs(execute.getDependencies().begin(),
+                                        execute.getDependencies().end());
+  functionInputs.insert(execute.getBodyOperands().begin(),
+                        execute.getBodyOperands().end());
+  getUsedValuesDefinedAbove(execute.getBodyRegion(), functionInputs);
 
   // Collect types for the outlined function inputs and outputs.
   auto typesRange = llvm::map_range(
@@ -257,7 +272,8 @@ outlineExecuteOp(SymbolTable &symbolTable, ExecuteOp execute) {
 
   // TODO: Derive outlined function name from the parent FuncOp (support
   // multiple nested async.execute operations).
-  FuncOp func = FuncOp::create(loc, kAsyncFnPrefix, funcType, funcAttrs);
+  func::FuncOp func =
+      func::FuncOp::create(loc, kAsyncFnPrefix, funcType, funcAttrs);
   symbolTable.insert(func);
 
   SymbolTable::setSymbolVisibility(func, SymbolTable::Visibility::Private);
@@ -265,8 +281,8 @@ outlineExecuteOp(SymbolTable &symbolTable, ExecuteOp execute) {
 
   // Prepare for coroutine conversion by creating the body of the function.
   {
-    size_t numDependencies = execute.dependencies().size();
-    size_t numOperands = execute.operands().size();
+    size_t numDependencies = execute.getDependencies().size();
+    size_t numOperands = execute.getBodyOperands().size();
 
     // Await on all dependencies before starting to execute the body region.
     for (size_t i = 0; i < numDependencies; ++i)
@@ -276,18 +292,18 @@ outlineExecuteOp(SymbolTable &symbolTable, ExecuteOp execute) {
     SmallVector<Value, 4> unwrappedOperands(numOperands);
     for (size_t i = 0; i < numOperands; ++i) {
       Value operand = func.getArgument(numDependencies + i);
-      unwrappedOperands[i] = builder.create<AwaitOp>(loc, operand).result();
+      unwrappedOperands[i] = builder.create<AwaitOp>(loc, operand).getResult();
     }
 
     // Map from function inputs defined above the execute op to the function
     // arguments.
     BlockAndValueMapping valueMapping;
     valueMapping.map(functionInputs, func.getArguments());
-    valueMapping.map(execute.body().getArguments(), unwrappedOperands);
+    valueMapping.map(execute.getBodyRegion().getArguments(), unwrappedOperands);
 
     // Clone all operations from the execute operation body into the outlined
     // function body.
-    for (Operation &op : execute.body().getOps())
+    for (Operation &op : execute.getBodyRegion().getOps())
       builder.clone(op, valueMapping);
   }
 
@@ -298,7 +314,7 @@ outlineExecuteOp(SymbolTable &symbolTable, ExecuteOp execute) {
   // Async resume operation (execution will be resumed in a thread managed by
   // the async runtime).
   {
-    BranchOp branch = cast<BranchOp>(coro.entry->getTerminator());
+    cf::BranchOp branch = cast<cf::BranchOp>(coro.entry->getTerminator());
     builder.setInsertionPointToEnd(coro.entry);
 
     // Save the coroutine state: async.coro.save
@@ -310,7 +326,7 @@ outlineExecuteOp(SymbolTable &symbolTable, ExecuteOp execute) {
     builder.create<RuntimeResumeOp>(coro.coroHandle);
 
     // Add async.coro.suspend as a suspended block terminator.
-    builder.create<CoroSuspendOp>(coroSaveOp.state(), coro.suspend,
+    builder.create<CoroSuspendOp>(coroSaveOp.getState(), coro.suspend,
                                   branch.getDest(), coro.cleanup);
 
     branch.erase();
@@ -319,7 +335,7 @@ outlineExecuteOp(SymbolTable &symbolTable, ExecuteOp execute) {
   // Replace the original `async.execute` with a call to outlined function.
   {
     ImplicitLocOpBuilder callBuilder(loc, execute);
-    auto callOutlinedFunc = callBuilder.create<CallOp>(
+    auto callOutlinedFunc = callBuilder.create<func::CallOp>(
         func.getName(), execute.getResultTypes(), functionInputs.getArrayRef());
     execute.replaceAllUsesWith(callOutlinedFunc.getResults());
     execute.erase();
@@ -377,8 +393,9 @@ class AwaitOpLoweringBase : public OpConversionPattern<AwaitType> {
   using AwaitAdaptor = typename AwaitType::Adaptor;
 
 public:
-  AwaitOpLoweringBase(MLIRContext *ctx,
-                      llvm::DenseMap<FuncOp, CoroMachinery> &outlinedFunctions)
+  AwaitOpLoweringBase(
+      MLIRContext *ctx,
+      llvm::DenseMap<func::FuncOp, CoroMachinery> &outlinedFunctions)
       : OpConversionPattern<AwaitType>(ctx),
         outlinedFunctions(outlinedFunctions) {}
 
@@ -387,16 +404,16 @@ public:
                   ConversionPatternRewriter &rewriter) const override {
     // We can only await on one the `AwaitableType` (for `await` it can be
     // a `token` or a `value`, for `await_all` it must be a `group`).
-    if (!op.operand().getType().template isa<AwaitableType>())
+    if (!op.getOperand().getType().template isa<AwaitableType>())
       return rewriter.notifyMatchFailure(op, "unsupported awaitable type");
 
     // Check if await operation is inside the outlined coroutine function.
-    auto func = op->template getParentOfType<FuncOp>();
+    auto func = op->template getParentOfType<func::FuncOp>();
     auto outlined = outlinedFunctions.find(func);
     const bool isInCoroutine = outlined != outlinedFunctions.end();
 
     Location loc = op->getLoc();
-    Value operand = adaptor.operand();
+    Value operand = adaptor.getOperand();
 
     Type i1 = rewriter.getI1Type();
 
@@ -412,8 +429,8 @@ public:
           isError, builder.create<arith::ConstantOp>(
                        loc, i1, builder.getIntegerAttr(i1, 1)));
 
-      builder.create<AssertOp>(notError,
-                               "Awaited async operand is in error state");
+      builder.create<cf::AssertOp>(notError,
+                                   "Awaited async operand is in error state");
     }
 
     // Inside the coroutine we convert await operation into coroutine suspension
@@ -436,7 +453,7 @@ public:
 
       // Add async.coro.suspend as a suspended block terminator.
       builder.setInsertionPointToEnd(suspended);
-      builder.create<CoroSuspendOp>(coroSaveOp.state(), coro.suspend, resume,
+      builder.create<CoroSuspendOp>(coroSaveOp.getState(), coro.suspend, resume,
                                     coro.cleanup);
 
       // Split the resume block into error checking and continuation.
@@ -445,11 +462,11 @@ public:
       // Check if the awaited value is in the error state.
       builder.setInsertionPointToStart(resume);
       auto isError = builder.create<RuntimeIsErrorOp>(loc, i1, operand);
-      builder.create<CondBranchOp>(isError,
-                                   /*trueDest=*/setupSetErrorBlock(coro),
-                                   /*trueArgs=*/ArrayRef<Value>(),
-                                   /*falseDest=*/continuation,
-                                   /*falseArgs=*/ArrayRef<Value>());
+      builder.create<cf::CondBranchOp>(isError,
+                                       /*trueDest=*/setupSetErrorBlock(coro),
+                                       /*trueArgs=*/ArrayRef<Value>(),
+                                       /*falseDest=*/continuation,
+                                       /*falseArgs=*/ArrayRef<Value>());
 
       // Make sure that replacement value will be constructed in the
       // continuation block.
@@ -471,7 +488,7 @@ public:
   }
 
 private:
-  llvm::DenseMap<FuncOp, CoroMachinery> &outlinedFunctions;
+  llvm::DenseMap<func::FuncOp, CoroMachinery> &outlinedFunctions;
 };
 
 /// Lowering for `async.await` with a token operand.
@@ -516,7 +533,7 @@ class YieldOpLowering : public OpConversionPattern<async::YieldOp> {
 public:
   YieldOpLowering(
       MLIRContext *ctx,
-      const llvm::DenseMap<FuncOp, CoroMachinery> &outlinedFunctions)
+      const llvm::DenseMap<func::FuncOp, CoroMachinery> &outlinedFunctions)
       : OpConversionPattern<async::YieldOp>(ctx),
         outlinedFunctions(outlinedFunctions) {}
 
@@ -524,7 +541,7 @@ public:
   matchAndRewrite(async::YieldOp op, OpAdaptor adaptor,
                   ConversionPatternRewriter &rewriter) const override {
     // Check if yield operation is inside the async coroutine function.
-    auto func = op->template getParentOfType<FuncOp>();
+    auto func = op->template getParentOfType<func::FuncOp>();
     auto outlined = outlinedFunctions.find(func);
     if (outlined == outlinedFunctions.end())
       return rewriter.notifyMatchFailure(
@@ -549,25 +566,26 @@ public:
   }
 
 private:
-  const llvm::DenseMap<FuncOp, CoroMachinery> &outlinedFunctions;
+  const llvm::DenseMap<func::FuncOp, CoroMachinery> &outlinedFunctions;
 };
 
 //===----------------------------------------------------------------------===//
-// Convert std.assert operation to cond_br into `set_error` block.
+// Convert cf.assert operation to cf.cond_br into `set_error` block.
 //===----------------------------------------------------------------------===//
 
-class AssertOpLowering : public OpConversionPattern<AssertOp> {
+class AssertOpLowering : public OpConversionPattern<cf::AssertOp> {
 public:
-  AssertOpLowering(MLIRContext *ctx,
-                   llvm::DenseMap<FuncOp, CoroMachinery> &outlinedFunctions)
-      : OpConversionPattern<AssertOp>(ctx),
+  AssertOpLowering(
+      MLIRContext *ctx,
+      llvm::DenseMap<func::FuncOp, CoroMachinery> &outlinedFunctions)
+      : OpConversionPattern<cf::AssertOp>(ctx),
         outlinedFunctions(outlinedFunctions) {}
 
   LogicalResult
-  matchAndRewrite(AssertOp op, OpAdaptor adaptor,
+  matchAndRewrite(cf::AssertOp op, OpAdaptor adaptor,
                   ConversionPatternRewriter &rewriter) const override {
     // Check if assert operation is inside the async coroutine function.
-    auto func = op->template getParentOfType<FuncOp>();
+    auto func = op->template getParentOfType<func::FuncOp>();
     auto outlined = outlinedFunctions.find(func);
     if (outlined == outlinedFunctions.end())
       return rewriter.notifyMatchFailure(
@@ -578,18 +596,18 @@ public:
 
     Block *cont = rewriter.splitBlock(op->getBlock(), Block::iterator(op));
     rewriter.setInsertionPointToEnd(cont->getPrevNode());
-    rewriter.create<CondBranchOp>(loc, adaptor.getArg(),
-                                  /*trueDest=*/cont,
-                                  /*trueArgs=*/ArrayRef<Value>(),
-                                  /*falseDest=*/setupSetErrorBlock(coro),
-                                  /*falseArgs=*/ArrayRef<Value>());
+    rewriter.create<cf::CondBranchOp>(loc, adaptor.getArg(),
+                                      /*trueDest=*/cont,
+                                      /*trueArgs=*/ArrayRef<Value>(),
+                                      /*falseDest=*/setupSetErrorBlock(coro),
+                                      /*falseArgs=*/ArrayRef<Value>());
     rewriter.eraseOp(op);
 
     return success();
   }
 
 private:
-  llvm::DenseMap<FuncOp, CoroMachinery> &outlinedFunctions;
+  llvm::DenseMap<func::FuncOp, CoroMachinery> &outlinedFunctions;
 };
 
 //===----------------------------------------------------------------------===//
@@ -599,18 +617,19 @@ private:
 /// 2) Prepending the results with `async.token`.
 /// 3) Setting up coroutine blocks.
 /// 4) Rewriting return ops as yield op and branch op into the suspend block.
-static CoroMachinery rewriteFuncAsCoroutine(FuncOp func) {
+static CoroMachinery rewriteFuncAsCoroutine(func::FuncOp func) {
   auto *ctx = func->getContext();
   auto loc = func.getLoc();
   SmallVector<Type> resultTypes;
   resultTypes.reserve(func.getCallableResults().size());
   llvm::transform(func.getCallableResults(), std::back_inserter(resultTypes),
                   [](Type type) { return ValueType::get(type); });
-  func.setType(FunctionType::get(ctx, func.getType().getInputs(), resultTypes));
+  func.setType(
+      FunctionType::get(ctx, func.getFunctionType().getInputs(), resultTypes));
   func.insertResult(0, TokenType::get(ctx), {});
   for (Block &block : func.getBlocks()) {
     Operation *terminator = block.getTerminator();
-    if (auto returnOp = dyn_cast<ReturnOp>(*terminator)) {
+    if (auto returnOp = dyn_cast<func::ReturnOp>(*terminator)) {
       ImplicitLocOpBuilder builder(loc, returnOp);
       builder.create<YieldOp>(returnOp.getOperands());
       returnOp.erase();
@@ -623,10 +642,11 @@ static CoroMachinery rewriteFuncAsCoroutine(FuncOp func) {
 ///
 /// The invocation of this function is safe only when call ops are traversed in
 /// reverse order of how they appear in a single block. See `funcsToCoroutines`.
-static void rewriteCallsiteForCoroutine(CallOp oldCall, FuncOp func) {
+static void rewriteCallsiteForCoroutine(func::CallOp oldCall,
+                                        func::FuncOp func) {
   auto loc = func.getLoc();
   ImplicitLocOpBuilder callBuilder(loc, oldCall);
-  auto newCall = callBuilder.create<CallOp>(
+  auto newCall = callBuilder.create<func::CallOp>(
       func.getName(), func.getCallableResults(), oldCall.getArgOperands());
 
   // Await on the async token and all the value results and unwrap the latter.
@@ -635,32 +655,32 @@ static void rewriteCallsiteForCoroutine(CallOp oldCall, FuncOp func) {
   unwrappedResults.reserve(newCall->getResults().size() - 1);
   for (Value result : newCall.getResults().drop_front())
     unwrappedResults.push_back(
-        callBuilder.create<AwaitOp>(loc, result).result());
+        callBuilder.create<AwaitOp>(loc, result).getResult());
   // Careful, when result of a call is piped into another call this could lead
   // to a dangling pointer.
   oldCall.replaceAllUsesWith(unwrappedResults);
   oldCall.erase();
 }
 
-static bool isAllowedToBlock(FuncOp func) {
+static bool isAllowedToBlock(func::FuncOp func) {
   return !!func->getAttrOfType<UnitAttr>(AsyncDialect::kAllowedToBlockAttrName);
 }
 
-static LogicalResult
-funcsToCoroutines(ModuleOp module,
-                  llvm::DenseMap<FuncOp, CoroMachinery> &outlinedFunctions) {
+static LogicalResult funcsToCoroutines(
+    ModuleOp module,
+    llvm::DenseMap<func::FuncOp, CoroMachinery> &outlinedFunctions) {
   // The following code supports the general case when 2 functions mutually
   // recurse into each other. Because of this and that we are relying on
   // SymbolUserMap to find pointers to calling FuncOps, we cannot simply erase
   // a FuncOp while inserting an equivalent coroutine, because that could lead
   // to dangling pointers.
 
-  SmallVector<FuncOp> funcWorklist;
+  SmallVector<func::FuncOp> funcWorklist;
 
   // Careful, it's okay to add a func to the worklist multiple times if and only
   // if the loop processing the worklist will skip the functions that have
   // already been converted to coroutines.
-  auto addToWorklist = [&](FuncOp func) {
+  auto addToWorklist = [&](func::FuncOp func) {
     if (isAllowedToBlock(func))
       return;
     // N.B. To refactor this code into a separate pass the lookup in
@@ -669,8 +689,8 @@ funcsToCoroutines(ModuleOp module,
     // this dict between the passes is ugly.
     if (isAllowedToBlock(func) ||
         outlinedFunctions.find(func) == outlinedFunctions.end()) {
-      for (Operation &op : func.body().getOps()) {
-        if (dyn_cast<AwaitOp>(op) || dyn_cast<AwaitAllOp>(op)) {
+      for (Operation &op : func.getBody().getOps()) {
+        if (isa<AwaitOp, AwaitAllOp>(op)) {
           funcWorklist.push_back(func);
           break;
         }
@@ -679,7 +699,7 @@ funcsToCoroutines(ModuleOp module,
   };
 
   // Traverse in post-order collecting for each func op the await ops it has.
-  for (FuncOp func : module.getOps<FuncOp>())
+  for (func::FuncOp func : module.getOps<func::FuncOp>())
     addToWorklist(func);
 
   SymbolTableCollection symbolTable;
@@ -708,8 +728,8 @@ funcsToCoroutines(ModuleOp module,
     });
     // Rewrite the callsites to await on results of the newly created coroutine.
     for (Operation *op : users) {
-      if (CallOp call = dyn_cast<mlir::CallOp>(*op)) {
-        FuncOp caller = call->getParentOfType<FuncOp>();
+      if (func::CallOp call = dyn_cast<func::CallOp>(*op)) {
+        func::FuncOp caller = call->getParentOfType<func::FuncOp>();
         rewriteCallsiteForCoroutine(call, func); // Careful, erases the call op.
         addToWorklist(caller);
       } else {
@@ -727,7 +747,7 @@ void AsyncToAsyncRuntimePass::runOnOperation() {
   SymbolTable symbolTable(module);
 
   // Outline all `async.execute` body regions into async functions (coroutines).
-  llvm::DenseMap<FuncOp, CoroMachinery> outlinedFunctions;
+  llvm::DenseMap<func::FuncOp, CoroMachinery> outlinedFunctions;
 
   module.walk([&](ExecuteOp execute) {
     outlinedFunctions.insert(outlineExecuteOp(symbolTable, execute));
@@ -740,7 +760,7 @@ void AsyncToAsyncRuntimePass::runOnOperation() {
 
   // Returns true if operation is inside the coroutine.
   auto isInCoroutine = [&](Operation *op) -> bool {
-    auto parentFunc = op->getParentOfType<FuncOp>();
+    auto parentFunc = op->getParentOfType<func::FuncOp>();
     return outlinedFunctions.find(parentFunc) != outlinedFunctions.end();
   };
 
@@ -758,7 +778,7 @@ void AsyncToAsyncRuntimePass::runOnOperation() {
   // and we have to make sure that structured control flow operations with async
   // operations in nested regions will be converted to branch-based control flow
   // before we add the coroutine basic blocks.
-  populateLoopToStdConversionPatterns(asyncPatterns);
+  populateSCFToControlFlowConversionPatterns(asyncPatterns);
 
   // Async lowering does not use type converter because it must preserve all
   // types for async.runtime operations.
@@ -785,19 +805,20 @@ void AsyncToAsyncRuntimePass::runOnOperation() {
     });
     return !walkResult.wasInterrupted();
   });
-  runtimeTarget.addLegalOp<AssertOp, arith::XOrIOp, arith::ConstantOp,
-                           ConstantOp, BranchOp, CondBranchOp>();
+  runtimeTarget.addLegalOp<cf::AssertOp, arith::XOrIOp, arith::ConstantOp,
+                           func::ConstantOp, cf::BranchOp, cf::CondBranchOp>();
 
   // Assertions must be converted to runtime errors inside async functions.
-  runtimeTarget.addDynamicallyLegalOp<AssertOp>([&](AssertOp op) -> bool {
-    auto func = op->getParentOfType<FuncOp>();
-    return outlinedFunctions.find(func) == outlinedFunctions.end();
-  });
+  runtimeTarget.addDynamicallyLegalOp<cf::AssertOp>(
+      [&](cf::AssertOp op) -> bool {
+        auto func = op->getParentOfType<func::FuncOp>();
+        return outlinedFunctions.find(func) == outlinedFunctions.end();
+      });
 
   if (eliminateBlockingAwaitOps)
     runtimeTarget.addDynamicallyLegalOp<RuntimeAwaitOp>(
         [&](RuntimeAwaitOp op) -> bool {
-          return isAllowedToBlock(op->getParentOfType<FuncOp>());
+          return isAllowedToBlock(op->getParentOfType<func::FuncOp>());
         });
 
   if (failed(applyPartialConversion(module, runtimeTarget,
