@@ -51,10 +51,20 @@ static inline bool isFnEntryBlock(Block *block) {
 spirv::Deserializer::Deserializer(ArrayRef<uint32_t> binary,
                                   MLIRContext *context)
     : binary(binary), context(context), unknownLoc(UnknownLoc::get(context)),
-      module(createModuleOp()), opBuilder(module->getRegion()) {}
+      module(createModuleOp()), opBuilder(module->getRegion())
+#ifndef NDEBUG
+      ,
+      logger(llvm::dbgs())
+#endif
+{
+}
 
 LogicalResult spirv::Deserializer::deserialize() {
-  LLVM_DEBUG(llvm::dbgs() << "+++ starting deserialization +++\n");
+  LLVM_DEBUG({
+    logger.resetIndent();
+    logger.startLine()
+        << "//+++---------- start deserialization ----------+++//\n";
+  });
 
   if (failed(processHeader()))
     return failure();
@@ -83,7 +93,8 @@ LogicalResult spirv::Deserializer::deserialize() {
 
   attachVCETriple();
 
-  LLVM_DEBUG(llvm::dbgs() << "+++ completed deserialization +++\n");
+  LLVM_DEBUG(logger.startLine()
+             << "//+++-------- completed deserialization --------+++//\n");
   return success();
 }
 
@@ -205,10 +216,11 @@ spirv::Deserializer::processMemoryModel(ArrayRef<uint32_t> operands) {
 
   (*module)->setAttr(
       "addressing_model",
-      opBuilder.getI32IntegerAttr(llvm::bit_cast<int32_t>(operands.front())));
-  (*module)->setAttr(
-      "memory_model",
-      opBuilder.getI32IntegerAttr(llvm::bit_cast<int32_t>(operands.back())));
+      opBuilder.getAttr<spirv::AddressingModelAttr>(
+          static_cast<spirv::AddressingModel>(operands.front())));
+  (*module)->setAttr("memory_model",
+                     opBuilder.getAttr<spirv::MemoryModelAttr>(
+                         static_cast<spirv::MemoryModel>(operands.back())));
 
   return success();
 }
@@ -227,7 +239,7 @@ LogicalResult spirv::Deserializer::processDecoration(ArrayRef<uint32_t> words) {
     return emitError(unknownLoc, "invalid Decoration code : ") << words[1];
   }
   auto attrName = llvm::convertToSnakeFromCamelCase(decorationName);
-  auto symbol = opBuilder.getIdentifier(attrName);
+  auto symbol = opBuilder.getStringAttr(attrName);
   switch (static_cast<spirv::Decoration>(words[1])) {
   case spirv::Decoration::DescriptorSet:
   case spirv::Decoration::Binding:
@@ -267,8 +279,8 @@ LogicalResult spirv::Deserializer::processDecoration(ArrayRef<uint32_t> words) {
       return emitError(unknownLoc, "OpDecoration with ")
              << decorationName << "needs a single target <id>";
     }
-    // Block decoration does not affect spv.struct type, but is still stored for
-    // verification.
+    // Block decoration does not affect spirv.struct type, but is still stored
+    // for verification.
     // TODO: Update StructType to contain this information since
     // it is needed for many validation rules.
     decorations[words[0]].set(symbol, opBuilder.getUnitAttr());
@@ -340,7 +352,8 @@ spirv::Deserializer::processFunction(ArrayRef<uint32_t> operands) {
            << operands[0];
   }
 
-  if (funcMap.count(operands[1])) {
+  uint32_t fnID = operands[1];
+  if (funcMap.count(fnID)) {
     return emitError(unknownLoc, "duplicate function definition/declaration");
   }
 
@@ -363,15 +376,20 @@ spirv::Deserializer::processFunction(ArrayRef<uint32_t> operands) {
            << functionType << " and return type " << resultType << " specified";
   }
 
-  std::string fnName = getFunctionSymbol(operands[1]);
+  std::string fnName = getFunctionSymbol(fnID);
   auto funcOp = opBuilder.create<spirv::FuncOp>(
-      unknownLoc, fnName, functionType, fnControl.getValue());
-  curFunction = funcMap[operands[1]] = funcOp;
-  LLVM_DEBUG(llvm::dbgs() << "-- start function " << fnName << " (type = "
-                          << fnType << ", id = " << operands[1] << ") --\n");
+      unknownLoc, fnName, functionType, fnControl.value());
+  curFunction = funcMap[fnID] = funcOp;
   auto *entryBlock = funcOp.addEntryBlock();
-  LLVM_DEBUG(llvm::dbgs() << "[block] created entry block " << entryBlock
-                          << "\n");
+  LLVM_DEBUG({
+    logger.startLine()
+        << "//===-------------------------------------------===//\n";
+    logger.startLine() << "[fn] name: " << fnName << "\n";
+    logger.startLine() << "[fn] type: " << fnType << "\n";
+    logger.startLine() << "[fn] ID: " << fnID << "\n";
+    logger.startLine() << "[fn] entry block: " << entryBlock << "\n";
+    logger.indent();
+  });
 
   // Parse the op argument instructions
   if (functionType.getNumInputs()) {
@@ -403,7 +421,7 @@ spirv::Deserializer::processFunction(ArrayRef<uint32_t> operands) {
                << argDefinedType << " at argument " << i;
       }
       if (getValue(operands[1])) {
-        return emitError(unknownLoc, "duplicate definition of result <id> '")
+        return emitError(unknownLoc, "duplicate definition of result <id> ")
                << operands[1];
       }
       auto argValue = funcOp.getArgument(i);
@@ -428,9 +446,6 @@ spirv::Deserializer::processFunction(ArrayRef<uint32_t> operands) {
     return failure();
   }
   if (opcode == spirv::Opcode::OpFunctionEnd) {
-    LLVM_DEBUG(llvm::dbgs()
-               << "-- completed function '" << fnName << "' (type = " << fnType
-               << ", id = " << operands[1] << ") --\n");
     return processFunctionEnd(instOperands);
   }
   if (opcode != spirv::Opcode::OpLabel) {
@@ -457,8 +472,6 @@ spirv::Deserializer::processFunction(ArrayRef<uint32_t> operands) {
     return failure();
   }
 
-  LLVM_DEBUG(llvm::dbgs() << "-- completed function '" << fnName << "' (type = "
-                          << fnType << ", id = " << operands[1] << ") --\n");
   return processFunctionEnd(instOperands);
 }
 
@@ -470,7 +483,8 @@ spirv::Deserializer::processFunctionEnd(ArrayRef<uint32_t> operands) {
   }
 
   // Wire up block arguments from OpPhi instructions.
-  // Put all structured control flow in spv.mlir.selection/spv.mlir.loop ops.
+  // Put all structured control flow in spirv.mlir.selection/spirv.mlir.loop
+  // ops.
   if (failed(wireUpBlockArgument()) || failed(structurizeControlFlow())) {
     return failure();
   }
@@ -478,6 +492,11 @@ spirv::Deserializer::processFunctionEnd(ArrayRef<uint32_t> operands) {
   curBlock = nullptr;
   curFunction = llvm::None;
 
+  LLVM_DEBUG({
+    logger.unindent();
+    logger.startLine()
+        << "//===-------------------------------------------===//\n";
+  });
   return success();
 }
 
@@ -521,7 +540,7 @@ spirv::Deserializer::createSpecConstant(Location loc, uint32_t resultID,
                                                     defaultValue);
   if (decorations.count(resultID)) {
     for (auto attr : decorations[resultID].getAttrs())
-      op->setAttr(attr.first, attr.second);
+      op->setAttr(attr.getName(), attr.getValue());
   }
   specConstMap[resultID] = op;
   return op;
@@ -545,7 +564,7 @@ spirv::Deserializer::processGlobalVariable(ArrayRef<uint32_t> operands) {
   auto ptrType = type.dyn_cast<spirv::PointerType>();
   if (!ptrType) {
     return emitError(unknownLoc,
-                     "expected a result type <id> to be a spv.ptr, found : ")
+                     "expected a result type <id> to be a spirv.ptr, found : ")
            << type;
   }
   wordIndex++;
@@ -591,9 +610,8 @@ spirv::Deserializer::processGlobalVariable(ArrayRef<uint32_t> operands) {
 
   // Decorations.
   if (decorations.count(variableID)) {
-    for (auto attr : decorations[variableID].getAttrs()) {
-      varOp->setAttr(attr.first, attr.second);
-    }
+    for (auto attr : decorations[variableID].getAttrs())
+      varOp->setAttr(attr.getName(), attr.getValue());
   }
   globalVariableMap[variableID] = varOp;
   return success();
@@ -713,6 +731,8 @@ LogicalResult spirv::Deserializer::processType(spirv::Opcode opcode,
     return processCooperativeMatrixType(operands);
   case spirv::Opcode::OpTypeFunction:
     return processFunctionType(operands);
+  case spirv::Opcode::OpTypeJointMatrixINTEL:
+    return processJointMatrixType(operands);
   case spirv::Opcode::OpTypeImage:
     return processImageType(operands);
   case spirv::Opcode::OpTypeSampledImage:
@@ -867,7 +887,41 @@ spirv::Deserializer::processCooperativeMatrixType(ArrayRef<uint32_t> operands) {
   unsigned columns = getConstantInt(operands[4]).getInt();
 
   typeMap[operands[0]] = spirv::CooperativeMatrixNVType::get(
-      elementTy, scope.getValue(), rows, columns);
+      elementTy, scope.value(), rows, columns);
+  return success();
+}
+
+LogicalResult
+spirv::Deserializer::processJointMatrixType(ArrayRef<uint32_t> operands) {
+  if (operands.size() != 6) {
+    return emitError(unknownLoc, "OpTypeJointMatrix must have element "
+                                 "type and row x column parameters");
+  }
+
+  Type elementTy = getType(operands[1]);
+  if (!elementTy) {
+    return emitError(unknownLoc, "OpTypeJointMatrix references undefined <id> ")
+           << operands[1];
+  }
+
+  auto scope = spirv::symbolizeScope(getConstantInt(operands[5]).getInt());
+  if (!scope) {
+    return emitError(unknownLoc,
+                     "OpTypeJointMatrix references undefined scope <id> ")
+           << operands[5];
+  }
+  auto matrixLayout =
+      spirv::symbolizeMatrixLayout(getConstantInt(operands[4]).getInt());
+  if (!matrixLayout) {
+    return emitError(unknownLoc,
+                     "OpTypeJointMatrix references undefined scope <id> ")
+           << operands[4];
+  }
+  unsigned rows = getConstantInt(operands[2]).getInt();
+  unsigned columns = getConstantInt(operands[3]).getInt();
+
+  typeMap[operands[0]] = spirv::JointMatrixINTELType::get(
+      elementTy, scope.value(), rows, columns, matrixLayout.value());
   return success();
 }
 
@@ -1051,8 +1105,8 @@ spirv::Deserializer::processImageType(ArrayRef<uint32_t> operands) {
            << operands[7];
 
   typeMap[operands[0]] = spirv::ImageType::get(
-      elementTy, dim.getValue(), depthInfo.getValue(), arrayedInfo.getValue(),
-      samplingInfo.getValue(), samplerUseInfo.getValue(), format.getValue());
+      elementTy, dim.value(), depthInfo.value(), arrayedInfo.value(),
+      samplingInfo.value(), samplerUseInfo.value(), format.value());
   return success();
 }
 
@@ -1361,7 +1415,7 @@ Value spirv::Deserializer::materializeSpecConstantOperation(
   auto specConstOperationOp =
       opBuilder.create<spirv::SpecConstantOperationOp>(loc, resultType);
 
-  Region &body = specConstOperationOp.body();
+  Region &body = specConstOperationOp.getBody();
   // Move the new block into SpecConstantOperation's body.
   body.getBlocks().splice(body.end(), curBlock->getParent()->getBlocks(),
                           Region::iterator(enclosedBlock));
@@ -1408,17 +1462,17 @@ spirv::Deserializer::processConstantNull(ArrayRef<uint32_t> operands) {
 
 Block *spirv::Deserializer::getOrCreateBlock(uint32_t id) {
   if (auto *block = getBlock(id)) {
-    LLVM_DEBUG(llvm::dbgs() << "[block] got exiting block for id = " << id
-                            << " @ " << block << "\n");
+    LLVM_DEBUG(logger.startLine() << "[block] got exiting block for id = " << id
+                                  << " @ " << block << "\n");
     return block;
   }
 
   // We don't know where this block will be placed finally (in a
-  // spv.mlir.selection or spv.mlir.loop or function). Create it into the
+  // spirv.mlir.selection or spirv.mlir.loop or function). Create it into the
   // function for now and sort out the proper place later.
   auto *block = curFunction->addBlock();
-  LLVM_DEBUG(llvm::dbgs() << "[block] created block for id = " << id << " @ "
-                          << block << "\n");
+  LLVM_DEBUG(logger.startLine() << "[block] created block for id = " << id
+                                << " @ " << block << "\n");
   return blockMap[id] = block;
 }
 
@@ -1438,7 +1492,7 @@ LogicalResult spirv::Deserializer::processBranch(ArrayRef<uint32_t> operands) {
   // the same OpLine information.
   opBuilder.create<spirv::BranchOp>(loc, target);
 
-  (void)clearDebugLine();
+  clearDebugLine();
   return success();
 }
 
@@ -1472,7 +1526,7 @@ spirv::Deserializer::processBranchConditional(ArrayRef<uint32_t> operands) {
       /*trueArguments=*/ArrayRef<Value>(), falseBlock,
       /*falseArguments=*/ArrayRef<Value>(), weights);
 
-  (void)clearDebugLine();
+  clearDebugLine();
   return success();
 }
 
@@ -1488,7 +1542,8 @@ LogicalResult spirv::Deserializer::processLabel(ArrayRef<uint32_t> operands) {
   auto labelID = operands[0];
   // We may have forward declared this block.
   auto *block = getOrCreateBlock(labelID);
-  LLVM_DEBUG(llvm::dbgs() << "[block] populating block " << block << "\n");
+  LLVM_DEBUG(logger.startLine()
+             << "[block] populating block " << block << "\n");
   // If we have seen this block, make sure it was just a forward declaration.
   assert(block->empty() && "re-deserialize the same block!");
 
@@ -1563,11 +1618,11 @@ LogicalResult spirv::Deserializer::processPhi(ArrayRef<uint32_t> operands) {
 
   // Create a block argument for this OpPhi instruction.
   Type blockArgType = getType(operands[0]);
-  BlockArgument blockArg = curBlock->addArgument(blockArgType);
+  BlockArgument blockArg = curBlock->addArgument(blockArgType, unknownLoc);
   valueMap[operands[1]] = blockArg;
-  LLVM_DEBUG(llvm::dbgs() << "[phi] created block argument " << blockArg
-                          << " id = " << operands[1] << " of type "
-                          << blockArgType << '\n');
+  LLVM_DEBUG(logger.startLine()
+             << "[phi] created block argument " << blockArg
+             << " id = " << operands[1] << " of type " << blockArgType << "\n");
 
   // For each (value, predecessor) pair, insert the value to the predecessor's
   // blockPhiInfo entry so later we can fix the block argument there.
@@ -1576,8 +1631,8 @@ LogicalResult spirv::Deserializer::processPhi(ArrayRef<uint32_t> operands) {
     Block *predecessor = getOrCreateBlock(operands[i + 1]);
     std::pair<Block *, Block *> predecessorTargetPair{predecessor, curBlock};
     blockPhiInfo[predecessorTargetPair].push_back(value);
-    LLVM_DEBUG(llvm::dbgs() << "[phi] predecessor @ " << predecessor
-                            << " with arg id = " << value << '\n');
+    LLVM_DEBUG(logger.startLine() << "[phi] predecessor @ " << predecessor
+                                  << " with arg id = " << value << "\n");
   }
 
   return success();
@@ -1585,43 +1640,45 @@ LogicalResult spirv::Deserializer::processPhi(ArrayRef<uint32_t> operands) {
 
 namespace {
 /// A class for putting all blocks in a structured selection/loop in a
-/// spv.mlir.selection/spv.mlir.loop op.
+/// spirv.mlir.selection/spirv.mlir.loop op.
 class ControlFlowStructurizer {
 public:
-  /// Structurizes the loop at the given `headerBlock`.
-  ///
-  /// This method will create an spv.mlir.loop op in the `mergeBlock` and move
-  /// all blocks in the structured loop into the spv.mlir.loop's region. All
-  /// branches to the `headerBlock` will be redirected to the `mergeBlock`. This
-  /// method will also update `mergeInfo` by remapping all blocks inside to the
-  /// newly cloned ones inside structured control flow op's regions.
-  static LogicalResult structurize(Location loc, uint32_t control,
-                                   spirv::BlockMergeInfoMap &mergeInfo,
-                                   Block *headerBlock, Block *mergeBlock,
-                                   Block *continueBlock) {
-    return ControlFlowStructurizer(loc, control, mergeInfo, headerBlock,
-                                   mergeBlock, continueBlock)
-        .structurizeImpl();
-  }
-
-private:
+#ifndef NDEBUG
+  ControlFlowStructurizer(Location loc, uint32_t control,
+                          spirv::BlockMergeInfoMap &mergeInfo, Block *header,
+                          Block *merge, Block *cont,
+                          llvm::ScopedPrinter &logger)
+      : location(loc), control(control), blockMergeInfo(mergeInfo),
+        headerBlock(header), mergeBlock(merge), continueBlock(cont),
+        logger(logger) {}
+#else
   ControlFlowStructurizer(Location loc, uint32_t control,
                           spirv::BlockMergeInfoMap &mergeInfo, Block *header,
                           Block *merge, Block *cont)
       : location(loc), control(control), blockMergeInfo(mergeInfo),
         headerBlock(header), mergeBlock(merge), continueBlock(cont) {}
+#endif
 
-  /// Creates a new spv.mlir.selection op at the beginning of the `mergeBlock`.
+  /// Structurizes the loop at the given `headerBlock`.
+  ///
+  /// This method will create an spirv.mlir.loop op in the `mergeBlock` and move
+  /// all blocks in the structured loop into the spirv.mlir.loop's region. All
+  /// branches to the `headerBlock` will be redirected to the `mergeBlock`. This
+  /// method will also update `mergeInfo` by remapping all blocks inside to the
+  /// newly cloned ones inside structured control flow op's regions.
+  LogicalResult structurize();
+
+private:
+  /// Creates a new spirv.mlir.selection op at the beginning of the
+  /// `mergeBlock`.
   spirv::SelectionOp createSelectionOp(uint32_t selectionControl);
 
-  /// Creates a new spv.mlir.loop op at the beginning of the `mergeBlock`.
+  /// Creates a new spirv.mlir.loop op at the beginning of the `mergeBlock`.
   spirv::LoopOp createLoopOp(uint32_t loopControl);
 
   /// Collects all blocks reachable from `headerBlock` except `mergeBlock`.
   void collectBlocksInConstruct();
 
-  LogicalResult structurizeImpl();
-
   Location location;
   uint32_t control;
 
@@ -1629,9 +1686,14 @@ private:
 
   Block *headerBlock;
   Block *mergeBlock;
-  Block *continueBlock; // nullptr for spv.mlir.selection
+  Block *continueBlock; // nullptr for spirv.mlir.selection
 
   SetVector<Block *> constructBlocks;
+
+#ifndef NDEBUG
+  /// A logger used to emit information during the deserialzation process.
+  llvm::ScopedPrinter &logger;
+#endif
 };
 } // namespace
 
@@ -1675,7 +1737,7 @@ void ControlFlowStructurizer::collectBlocksInConstruct() {
   }
 }
 
-LogicalResult ControlFlowStructurizer::structurizeImpl() {
+LogicalResult ControlFlowStructurizer::structurize() {
   Operation *op = nullptr;
   bool isLoop = continueBlock != nullptr;
   if (isLoop) {
@@ -1721,19 +1783,21 @@ LogicalResult ControlFlowStructurizer::structurizeImpl() {
     // SelectionOp/LoopOp's region.
     auto *newBlock = builder.createBlock(&body.back());
     mapper.map(block, newBlock);
-    LLVM_DEBUG(llvm::dbgs() << "[cf] cloned block " << newBlock
-                            << " from block " << block << "\n");
+    LLVM_DEBUG(logger.startLine() << "[cf] cloned block " << newBlock
+                                  << " from block " << block << "\n");
     if (!isFnEntryBlock(block)) {
       for (BlockArgument blockArg : block->getArguments()) {
-        auto newArg = newBlock->addArgument(blockArg.getType());
+        auto newArg =
+            newBlock->addArgument(blockArg.getType(), blockArg.getLoc());
         mapper.map(blockArg, newArg);
-        LLVM_DEBUG(llvm::dbgs() << "[cf] remapped block argument " << blockArg
-                                << " to " << newArg << '\n');
+        LLVM_DEBUG(logger.startLine() << "[cf] remapped block argument "
+                                      << blockArg << " to " << newArg << "\n");
       }
     } else {
-      LLVM_DEBUG(llvm::dbgs()
+      LLVM_DEBUG(logger.startLine()
                  << "[cf] block " << block << " is a function entry block\n");
     }
+
     for (auto &op : *block)
       newBlock->push_back(op.clone(mapper));
   }
@@ -1747,9 +1811,8 @@ LogicalResult ControlFlowStructurizer::structurizeImpl() {
       if (Block *mappedOp = mapper.lookupOrNull(succOp.get()))
         succOp.set(mappedOp);
   };
-  for (auto &block : body) {
+  for (auto &block : body)
     block.walk(remapOperands);
-  }
 
   // We have created the SelectionOp/LoopOp and "moved" all blocks belonging to
   // the selection/loop construct into its region. Next we need to fix the
@@ -1759,17 +1822,26 @@ LogicalResult ControlFlowStructurizer::structurizeImpl() {
   // SelectionOp/LoopOp resides right now.
   headerBlock->replaceAllUsesWith(mergeBlock);
 
+  LLVM_DEBUG({
+    logger.startLine() << "[cf] after cloning and fixing references:\n";
+    headerBlock->getParentOp()->print(logger.getOStream());
+    logger.startLine() << "\n";
+  });
+
   if (isLoop) {
-    // The loop selection/loop header block may have block arguments. Since now
-    // we place the selection/loop op inside the old merge block, we need to
-    // make sure the old merge block has the same block argument list.
-    assert(mergeBlock->args_empty() && "OpPhi in loop merge block unsupported");
-    for (BlockArgument blockArg : headerBlock->getArguments()) {
-      mergeBlock->addArgument(blockArg.getType());
+    if (!mergeBlock->args_empty()) {
+      return mergeBlock->getParentOp()->emitError(
+          "OpPhi in loop merge block unsupported");
     }
 
-    // If the loop header block has block arguments, make sure the spv.branch op
-    // matches.
+    // The loop header block may have block arguments. Since now we place the
+    // loop op inside the old merge block, we need to make sure the old merge
+    // block has the same block argument list.
+    for (BlockArgument blockArg : headerBlock->getArguments())
+      mergeBlock->addArgument(blockArg.getType(), blockArg.getLoc());
+
+    // If the loop header block has block arguments, make sure the spirv.Branch
+    // op matches.
     SmallVector<Value, 4> blockArgs;
     if (!headerBlock->args_empty())
       blockArgs = {mergeBlock->args_begin(), mergeBlock->args_end()};
@@ -1783,12 +1855,25 @@ LogicalResult ControlFlowStructurizer::structurizeImpl() {
 
   // All the blocks cloned into the SelectionOp/LoopOp's region can now be
   // cleaned up.
-  LLVM_DEBUG(llvm::dbgs() << "[cf] cleaning up blocks after clone\n");
+  LLVM_DEBUG(logger.startLine() << "[cf] cleaning up blocks after clone\n");
   // First we need to drop all operands' references inside all blocks. This is
   // needed because we can have blocks referencing SSA values from one another.
   for (auto *block : constructBlocks)
     block->dropAllReferences();
 
+  // Check that whether some op in the to-be-erased blocks still has uses. Those
+  // uses come from blocks that won't be sinked into the SelectionOp/LoopOp's
+  // region. We cannot handle such cases given that once a value is sinked into
+  // the SelectionOp/LoopOp's region, there is no escape for it:
+  // SelectionOp/LooOp does not support yield values right now.
+  for (auto *block : constructBlocks) {
+    for (Operation &op : *block)
+      if (!op.use_empty())
+        return op.emitOpError(
+            "failed control flow structurization: it has uses outside of the "
+            "enclosing selection/loop construct");
+  }
+
   // Then erase all old blocks.
   for (auto *block : constructBlocks) {
     // We've cloned all blocks belonging to this construct into the structured
@@ -1796,26 +1881,31 @@ LogicalResult ControlFlowStructurizer::structurizeImpl() {
     // selection/loop. If so, they will be recorded within blockMergeInfo.
     // We need to update the pointers there to the newly remapped ones so we can
     // continue structurizing them later.
-    // TODO: The asserts in the following assumes input SPIR-V blob
-    // forms correctly nested selection/loop constructs. We should relax this
-    // and support error cases better.
+    // TODO: The asserts in the following assumes input SPIR-V blob forms
+    // correctly nested selection/loop constructs. We should relax this and
+    // support error cases better.
     auto it = blockMergeInfo.find(block);
     if (it != blockMergeInfo.end()) {
+      // Use the original location for nested selection/loop ops.
+      Location loc = it->second.loc;
+
       Block *newHeader = mapper.lookupOrNull(block);
-      assert(newHeader && "nested loop header block should be remapped!");
+      if (!newHeader)
+        return emitError(loc, "failed control flow structurization: nested "
+                              "loop header block should be remapped!");
 
       Block *newContinue = it->second.continueBlock;
       if (newContinue) {
         newContinue = mapper.lookupOrNull(newContinue);
-        assert(newContinue && "nested loop continue block should be remapped!");
+        if (!newContinue)
+          return emitError(loc, "failed control flow structurization: nested "
+                                "loop continue block should be remapped!");
       }
 
       Block *newMerge = it->second.mergeBlock;
       if (Block *mappedTo = mapper.lookupOrNull(newMerge))
         newMerge = mappedTo;
 
-      // Keep original location for nested selection/loop ops.
-      Location loc = it->second.loc;
       // The iterator should be erased before adding a new entry into
       // blockMergeInfo to avoid iterator invalidation.
       blockMergeInfo.erase(it);
@@ -1828,29 +1918,33 @@ LogicalResult ControlFlowStructurizer::structurizeImpl() {
     // flow, we cannot just simply erase it because it may contain arguments
     // matching the function signature and used by the cloned blocks.
     if (isFnEntryBlock(block)) {
-      LLVM_DEBUG(llvm::dbgs() << "[cf] changing entry block " << block
-                              << " to only contain a spv.Branch op\n");
+      LLVM_DEBUG(logger.startLine() << "[cf] changing entry block " << block
+                                    << " to only contain a spirv.Branch op\n");
       // Still keep the function entry block for the potential block arguments,
       // but replace all ops inside with a branch to the merge block.
       block->clear();
       builder.setInsertionPointToEnd(block);
       builder.create<spirv::BranchOp>(location, mergeBlock);
     } else {
-      LLVM_DEBUG(llvm::dbgs() << "[cf] erasing block " << block << "\n");
+      LLVM_DEBUG(logger.startLine() << "[cf] erasing block " << block << "\n");
       block->erase();
     }
   }
 
-  LLVM_DEBUG(
-      llvm::dbgs() << "[cf] after structurizing construct with header block "
-                   << headerBlock << ":\n"
-                   << *op << '\n');
+  LLVM_DEBUG(logger.startLine()
+             << "[cf] after structurizing construct with header block "
+             << headerBlock << ":\n"
+             << *op << "\n");
 
   return success();
 }
 
 LogicalResult spirv::Deserializer::wireUpBlockArgument() {
-  LLVM_DEBUG(llvm::dbgs() << "[phi] start wiring up block arguments\n");
+  LLVM_DEBUG({
+    logger.startLine()
+        << "//----- [phi] start wiring up block arguments -----//\n";
+    logger.indent();
+  });
 
   OpBuilder::InsertionGuard guard(opBuilder);
 
@@ -1858,10 +1952,12 @@ LogicalResult spirv::Deserializer::wireUpBlockArgument() {
     Block *block = info.first.first;
     Block *target = info.first.second;
     const BlockPhiInfo &phiInfo = info.second;
-    LLVM_DEBUG(llvm::dbgs() << "[phi] block " << block << "\n");
-    LLVM_DEBUG(llvm::dbgs() << "[phi] before creating block argument:\n");
-    LLVM_DEBUG(block->getParentOp()->print(llvm::dbgs()));
-    LLVM_DEBUG(llvm::dbgs() << '\n');
+    LLVM_DEBUG({
+      logger.startLine() << "[phi] block " << block << "\n";
+      logger.startLine() << "[phi] before creating block argument:\n";
+      block->getParentOp()->print(logger.getOStream());
+      logger.startLine() << "\n";
+    });
 
     // Set insertion point to before this block's terminator early because we
     // may materialize ops via getValue() call.
@@ -1873,8 +1969,8 @@ LogicalResult spirv::Deserializer::wireUpBlockArgument() {
     for (uint32_t valueId : phiInfo) {
       if (Value value = getValue(valueId)) {
         blockArgs.push_back(value);
-        LLVM_DEBUG(llvm::dbgs() << "[phi] block argument " << value
-                                << " id = " << valueId << '\n');
+        LLVM_DEBUG(logger.startLine() << "[phi] block argument " << value
+                                      << " id = " << valueId << "\n");
       } else {
         return emitError(unknownLoc, "OpPhi references undefined value!");
       }
@@ -1889,17 +1985,17 @@ LogicalResult spirv::Deserializer::wireUpBlockArgument() {
       assert((branchCondOp.getTrueBlock() == target ||
               branchCondOp.getFalseBlock() == target) &&
              "expected target to be either the true or false target");
-      if (target == branchCondOp.trueTarget())
+      if (target == branchCondOp.getTrueTarget())
         opBuilder.create<spirv::BranchConditionalOp>(
-            branchCondOp.getLoc(), branchCondOp.condition(), blockArgs,
+            branchCondOp.getLoc(), branchCondOp.getCondition(), blockArgs,
             branchCondOp.getFalseBlockArguments(),
-            branchCondOp.branch_weightsAttr(), branchCondOp.trueTarget(),
-            branchCondOp.falseTarget());
+            branchCondOp.getBranchWeightsAttr(), branchCondOp.getTrueTarget(),
+            branchCondOp.getFalseTarget());
       else
         opBuilder.create<spirv::BranchConditionalOp>(
-            branchCondOp.getLoc(), branchCondOp.condition(),
+            branchCondOp.getLoc(), branchCondOp.getCondition(),
             branchCondOp.getTrueBlockArguments(), blockArgs,
-            branchCondOp.branch_weightsAttr(), branchCondOp.getTrueBlock(),
+            branchCondOp.getBranchWeightsAttr(), branchCondOp.getTrueBlock(),
             branchCondOp.getFalseBlock());
 
       branchCondOp.erase();
@@ -1907,49 +2003,75 @@ LogicalResult spirv::Deserializer::wireUpBlockArgument() {
       return emitError(unknownLoc, "unimplemented terminator for Phi creation");
     }
 
-    LLVM_DEBUG(llvm::dbgs() << "[phi] after creating block argument:\n");
-    LLVM_DEBUG(block->getParentOp()->print(llvm::dbgs()));
-    LLVM_DEBUG(llvm::dbgs() << '\n');
+    LLVM_DEBUG({
+      logger.startLine() << "[phi] after creating block argument:\n";
+      block->getParentOp()->print(logger.getOStream());
+      logger.startLine() << "\n";
+    });
   }
   blockPhiInfo.clear();
 
-  LLVM_DEBUG(llvm::dbgs() << "[phi] completed wiring up block arguments\n");
+  LLVM_DEBUG({
+    logger.unindent();
+    logger.startLine()
+        << "//--- [phi] completed wiring up block arguments ---//\n";
+  });
   return success();
 }
 
 LogicalResult spirv::Deserializer::structurizeControlFlow() {
-  LLVM_DEBUG(llvm::dbgs() << "[cf] start structurizing control flow\n");
+  LLVM_DEBUG({
+    logger.startLine()
+        << "//----- [cf] start structurizing control flow -----//\n";
+    logger.indent();
+  });
 
   while (!blockMergeInfo.empty()) {
     Block *headerBlock = blockMergeInfo.begin()->first;
     BlockMergeInfo mergeInfo = blockMergeInfo.begin()->second;
 
-    LLVM_DEBUG(llvm::dbgs() << "[cf] header block " << headerBlock << ":\n");
-    LLVM_DEBUG(headerBlock->print(llvm::dbgs()));
+    LLVM_DEBUG({
+      logger.startLine() << "[cf] header block " << headerBlock << ":\n";
+      headerBlock->print(logger.getOStream());
+      logger.startLine() << "\n";
+    });
 
     auto *mergeBlock = mergeInfo.mergeBlock;
     assert(mergeBlock && "merge block cannot be nullptr");
     if (!mergeBlock->args_empty())
       return emitError(unknownLoc, "OpPhi in loop merge block unimplemented");
-    LLVM_DEBUG(llvm::dbgs() << "[cf] merge block " << mergeBlock << ":\n");
-    LLVM_DEBUG(mergeBlock->print(llvm::dbgs()));
+    LLVM_DEBUG({
+      logger.startLine() << "[cf] merge block " << mergeBlock << ":\n";
+      mergeBlock->print(logger.getOStream());
+      logger.startLine() << "\n";
+    });
 
     auto *continueBlock = mergeInfo.continueBlock;
-    if (continueBlock) {
-      LLVM_DEBUG(llvm::dbgs()
-                 << "[cf] continue block " << continueBlock << ":\n");
-      LLVM_DEBUG(continueBlock->print(llvm::dbgs()));
-    }
+    LLVM_DEBUG(if (continueBlock) {
+      logger.startLine() << "[cf] continue block " << continueBlock << ":\n";
+      continueBlock->print(logger.getOStream());
+      logger.startLine() << "\n";
+    });
     // Erase this case before calling into structurizer, who will update
     // blockMergeInfo.
     blockMergeInfo.erase(blockMergeInfo.begin());
-    if (failed(ControlFlowStructurizer::structurize(
-            mergeInfo.loc, mergeInfo.control, blockMergeInfo, headerBlock,
-            mergeBlock, continueBlock)))
+    ControlFlowStructurizer structurizer(mergeInfo.loc, mergeInfo.control,
+                                         blockMergeInfo, headerBlock,
+                                         mergeBlock, continueBlock
+#ifndef NDEBUG
+                                         ,
+                                         logger
+#endif
+    );
+    if (failed(structurizer.structurize()))
       return failure();
   }
 
-  LLVM_DEBUG(llvm::dbgs() << "[cf] completed structurizing control flow\n");
+  LLVM_DEBUG({
+    logger.unindent();
+    logger.startLine()
+        << "//--- [cf] completed structurizing control flow ---//\n";
+  });
   return success();
 }
 
@@ -1964,8 +2086,8 @@ Location spirv::Deserializer::createFileLineColLoc(OpBuilder opBuilder) {
   auto fileName = debugInfoMap.lookup(debugLine->fileID).str();
   if (fileName.empty())
     fileName = "<unknown>";
-  return FileLineColLoc::get(opBuilder.getIdentifier(fileName), debugLine->line,
-                             debugLine->col);
+  return FileLineColLoc::get(opBuilder.getStringAttr(fileName), debugLine->line,
+                             debugLine->column);
 }
 
 LogicalResult
@@ -1977,14 +2099,11 @@ spirv::Deserializer::processDebugLine(ArrayRef<uint32_t> operands) {
   // OpNoLine instruction."
   if (operands.size() != 3)
     return emitError(unknownLoc, "OpLine must have 3 operands");
-  debugLine = DebugLine(operands[0], operands[1], operands[2]);
+  debugLine = DebugLine{operands[0], operands[1], operands[2]};
   return success();
 }
 
-LogicalResult spirv::Deserializer::clearDebugLine() {
-  debugLine = llvm::None;
-  return success();
-}
+void spirv::Deserializer::clearDebugLine() { debugLine = llvm::None; }
 
 LogicalResult
 spirv::Deserializer::processDebugString(ArrayRef<uint32_t> operands) {
