@@ -2,19 +2,20 @@
 #  See https://llvm.org/LICENSE.txt for license information.
 #  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
 
-from typing import Dict, List, Sequence, Tuple, Union
+from typing import Callable, Dict, List, Sequence, Tuple, Union
 
 from .....ir import *
-from ....._mlir_libs._mlir.dialects.linalg import fill_builtin_region
 
+from .... import func
 from .... import linalg
-from .... import std
 from .... import math
 from .... import arith
+from .... import complex
 from ...._ods_common import get_op_result_or_value as _get_op_result_or_value, get_op_results_or_values as _get_op_results_or_values
 
 from .scalar_expr import *
 from .config import *
+from .comprehension import *
 import numpy as np
 
 __all__ = [
@@ -23,8 +24,10 @@ __all__ = [
     "ValueList",
 ]
 
+# Type aliases.
 ValueList = Union[Sequence[Value], OpResultList]
 
+
 def isa(cls: Type, ty: Type):
   try:
     cls(ty)
@@ -35,17 +38,29 @@ def isa(cls: Type, ty: Type):
 
 def prepare_common_structured_op(op_config: LinalgStructuredOpConfig,
                                  *ins: Value, outs: ValueList,
-                                 **attrs: Sequence[int]):
+                                 **attrs: Union[Sequence[int], TypeFnType]):
   all_arg_defs = op_config.ordered_operands
-  in_arg_defs = [arg for arg in all_arg_defs if arg.usage == "InputOperand"]
-  out_arg_defs = [arg for arg in all_arg_defs if arg.usage == "OutputOperand"]
-  attr_arg_defs = [arg for arg in all_arg_defs if arg.usage == "IndexAttribute"]
+  in_arg_defs = [
+      d for d in all_arg_defs
+      if d.kind in [OperandKind.SCALAR, OperandKind.INPUT_TENSOR]
+  ]
+  out_arg_defs = [
+      d for d in all_arg_defs if d.kind == OperandKind.OUTPUT_TENSOR
+  ]
+  index_attr_arg_defs = [
+      d for d in all_arg_defs if d.kind == OperandKind.INDEX_ATTR
+  ]
+  fn_attr_arg_defs = [
+      d for d in all_arg_defs if d.kind in [
+          OperandKind.UNARY_FN_ATTR, OperandKind.BINARY_FN_ATTR,
+          OperandKind.TYPE_FN_ATTR
+      ]
+  ]
 
   # Verify outs is a sequence or a list of results.
   if not isinstance(outs, (Sequence, OpResultList)):
-    raise ValueError(
-        f"Expected named argument outs to have type Sequence or OpResultLis but got {type(outs)}"
-    )
+    raise ValueError(f"Expected named argument outs to have type Sequence or "
+                     f"OpResultLis but got {type(outs)}")
 
   # Arity validation.
   if len(ins) != len(in_arg_defs):
@@ -55,21 +70,22 @@ def prepare_common_structured_op(op_config: LinalgStructuredOpConfig,
     raise ValueError(f"Expected {len(out_arg_defs)} outputs but got "
                      f"{len(outs)} for {op_config}")
 
-  # Compute a replacement list for all attribute symbols.
+  # Compute a replacement list for all index attribute symbols.
   expressions = []  # type: Sequence[AffineExpr]
   replacements = []  # type: Sequence[AffineExpr]
-  for attr in attr_arg_defs:
-    if attr.name not in attrs:
-      raise ValueError(f"Expected named argument for the attribute {attr.name}")
-    attribute_values = attrs.get(attr.name)
-    if not all(isinstance(value, int) for value in attribute_values):
-      raise ValueError(f"Attribute {attr.name} needs to be of type "
-                       f"Sequence[int] but got {type(attribute_values)}")
-    results = attr.attribute_map.results  # type: AffineExprList
-    if len(attribute_values) != len(results):
-      raise ValueError(f"Attribute {attr.name} has length {len(results)} "
-                       f"but got {len(attribute_values)} values")
-    for expr, value in zip(results, attribute_values):
+  for index_attr in index_attr_arg_defs:
+    index_attr_vals = index_attr.operand_def.default_indices
+    if index_attr.name in attrs:
+      index_attr_vals = attrs.get(index_attr.name)
+    assert index_attr_vals, "Index attribute has no value"
+    if not all(isinstance(value, int) for value in index_attr_vals):
+      raise ValueError(f"Attribute {index_attr.name} needs to be of type "
+                       f"Sequence[int] but got {type(index_attr_vals)}")
+    results = index_attr.index_attr_map.results  # type: AffineExprList
+    if len(index_attr_vals) != len(results):
+      raise ValueError(f"Attribute {index_attr.name} has length {len(results)} "
+                       f"but got {len(index_attr_vals)} values")
+    for expr, value in zip(results, index_attr_vals):
       expressions.append(expr)
       replacements.append(AffineConstantExpr.get(value))
 
@@ -114,24 +130,72 @@ def prepare_common_structured_op(op_config: LinalgStructuredOpConfig,
   iterator_types_attr = ArrayAttr.get(
       [StringAttr.get(s) for s in op_config.iterator_types])
 
-  # Compute a dictionary storing all index attributes.
-  index_attributes = {}  # type: Dict[str, DenseElementAttr]
-  for attr in attr_arg_defs:
-    attribute_values = attrs.get(attr.name)
-    array = np.array(attribute_values, dtype=np.int64)
-    index_attributes[attr.name] = DenseElementsAttr.get(array)
+  # Compute the index attributes used when emitting a named structured op.
+  index_attrs = {}  # type: Dict[str, DenseElementAttr]
+  for index_attr in index_attr_arg_defs:
+    index_attr_vals = attrs.get(index_attr.name)
+    # Only forward attributes set to a non-default value.
+    if index_attr_vals:
+      array = np.array(index_attr_vals, dtype=np.int64)
+      index_attrs[index_attr.name] = DenseElementsAttr.get(array)
+
+  # Compute the function attribute mapping.
+  fn_attr_mapping = {}
+  for fn_attr in fn_attr_arg_defs:
+    attr_val = fn_attr.operand_def.default_fn
+    attr_kind = fn_attr.kind
+    if fn_attr.name in attrs:
+      fn = attrs.get(fn_attr.name)
+      if attr_kind == OperandKind.UNARY_FN_ATTR:
+        if not isinstance(fn, UnaryFnType):
+          raise ValueError(f"Attribute {fn_attr.name} needs to be of type "
+                           f"UnaryFnType but got {type(attr_val)}")
+      elif attr_kind == OperandKind.BINARY_FN_ATTR:
+        if not isinstance(fn, BinaryFnType):
+          raise ValueError(f"Attribute {fn_attr.name} needs to be of type "
+                           f"BinaryFnType but got {type(attr_val)}")
+      else:
+        if not isinstance(fn, TypeFnType):
+          raise ValueError(f"Attribute {fn_attr.name} needs to be of type "
+                           f"TypeFnType but got {type(attr_val)}")
+      attr_val = fn.fn_name
+    assert attr_val, "Function attribute has no value"
+    fn_attr_mapping[fn_attr.name] = (attr_val, attr_kind)
 
   return (all_arg_defs, in_arg_defs, out_arg_defs, outs, result_types,
-          type_mapping, indexing_maps_attr, iterator_types_attr,
-          index_attributes, block_arg_types)
+          type_mapping, indexing_maps_attr, iterator_types_attr, index_attrs,
+          fn_attr_mapping, block_arg_types)
 
 
 def emit_generic_structured_op(op_config: LinalgStructuredOpConfig, *ins: Value,
                                outs: ValueList, **attrs: Sequence[int]):
   all_arg_defs, in_arg_defs, out_arg_defs, outs, result_types, type_mapping, \
-  indexing_maps_attr, iterator_types_attr, index_attributes, block_arg_types = \
+  indexing_maps_attr, iterator_types_attr, index_attrs, fn_attr_mapping, \
+  block_arg_types = \
      prepare_common_structured_op(op_config, *ins, outs = outs, **attrs)
 
+  # An operation that accesses only scalars and scalar/rank zero tensors is
+  # rank polymorhpic. We implement rank polymorphism by generating different
+  # indexing maps and iterators that match the rank of the first output tensor.
+  # An operation is rank polymorphic if the iteration domain has rank zero.
+  if not iterator_types_attr:
+    rank = ShapedType(outs[0].type).rank
+    iterator_types_attr = ArrayAttr.get([StringAttr.get("parallel")] * rank)
+    scalar_map = AffineMap.get(rank, 0, [])
+    tensor_map = AffineMap.get_identity(rank)
+    indexing_maps = []
+    for arg_def in all_arg_defs:
+      if arg_def.operand_def.kind == OperandKind.SCALAR:
+        indexing_maps.append(scalar_map)
+      if arg_def.operand_def.is_tensor():
+        idx = arg_def.operand_def.registered_index
+        if idx < len(ins) and ShapedType(ins[idx].type).rank == 0:
+          indexing_maps.append(scalar_map)
+        else:
+          indexing_maps.append(tensor_map)
+    indexing_maps_attr = ArrayAttr.get(
+        [AffineMapAttr.get(am) for am in indexing_maps])
+
   generic_op = linalg.GenericOp(
       result_tensors=result_types,
       inputs=ins,
@@ -146,7 +210,8 @@ def emit_generic_structured_op(op_config: LinalgStructuredOpConfig, *ins: Value,
   block = generic_op.regions[0].blocks.append(*block_arg_types)
   block_arg_mapping = dict(zip(block_arg_names, block.arguments))
   with InsertionPoint(block):
-    body_builder = _BodyBuilder(type_mapping, block_arg_mapping)
+    body_builder = _BodyBuilder(type_mapping, block_arg_mapping,
+                                fn_attr_mapping)
     for assignment in op_config.assignments:
       body_builder.assign(assignment)
     body_builder.yield_outputs(*_get_operand_def_names(*out_arg_defs))
@@ -161,7 +226,8 @@ def emit_named_structured_op(op_config: LinalgStructuredOpConfig, op_name: str,
                              op_class_name: str, *ins: Value, outs: ValueList,
                              **attrs: Sequence[int]):
   all_arg_defs, in_arg_defs, out_arg_defs, outs, result_types, type_mapping, \
-  indexing_maps_attr, iterator_types_attr, index_attributes, block_arg_types = \
+  indexing_maps_attr, iterator_types_attr, index_attrs, fn_attr_mapping, \
+  block_arg_types = \
      prepare_common_structured_op(op_config, *ins, outs = outs, **attrs)
 
   # If we get here, there must exist a builtin class `op_class_name`.
@@ -172,20 +238,20 @@ def emit_named_structured_op(op_config: LinalgStructuredOpConfig, op_name: str,
     raise NotImplementedError(
         f"Unknown named op_name / op_class_name: {op_name} / {op_class_name}")
 
+  # Set the index attributes used to compute the indexing maps.
   named_op = getattr(linalg, op_class_name)(ins, outs, result_types)
-  linalgDialect = ctx.get_dialect_descriptor("linalg")
-  fill_builtin_region(linalgDialect, named_op.operation)
-  # Note: mlir-linalg-ods-yaml-gen.cpp uses a special linalg.memoized_indexing_maps
-  # attribute that the non-yaml path does not. The non-yaml path hardcodes the
-  # indexing_maps in C++ directly.
-  named_op.operation.attributes[
-      "linalg.memoized_indexing_maps"] = indexing_maps_attr
-  # iterator_types are hardcoded in C++ both in the yaml and non-yaml path.
-
-  # Additionally set all named attributes.
-  for name, value in index_attributes.items():
+  for name, value in index_attrs.items():
     named_op.operation.attributes[name] = value
 
+  # Compute the function attributes by combining operand kind and function name.
+  for name, (fn_name, kind) in fn_attr_mapping.items():
+    assert kind.name.lower().endswith("_attr")
+    enum_name = kind.name.lower()[:-5]
+    named_op.operation.attributes[name] = Attribute.parse(
+        f"#linalg.{enum_name}<{fn_name}>")
+
+  linalg.fill_builtin_region(named_op.operation)
+
   if len(result_types) == 1:
     return named_op.result
   else:
@@ -196,9 +262,11 @@ class _BodyBuilder:
   """Constructs a structured op body by evaluating assignments."""
 
   def __init__(self, type_mapping: Dict[str, Type],
-               block_arg_mapping: Dict[str, Value]):
+               block_arg_mapping: Dict[str, Value], fn_attr_mapping: Dict[str,
+                                                                          str]):
     self.type_mapping = type_mapping
     self.block_arg_mapping = block_arg_mapping
+    self.fn_attr_mapping = fn_attr_mapping
     self.yield_mapping = dict()  # type: Dict[str, Value]
 
   def assign(self, assignment: ScalarAssign):
@@ -222,25 +290,41 @@ class _BodyBuilder:
       dim_attr = IntegerAttr.get(
           IntegerType.get_signless(64), expr.scalar_index.dim)
       return linalg.IndexOp(dim_attr).result
-    elif expr.scalar_apply:
-      try:
-        fn = getattr(self, f"_eval_{expr.scalar_apply.fn_name}")
-      except AttributeError:
-        raise ValueError(
-            f"Function '{expr.scalar_apply.fn_name}' is not a known "
-            "scalar body function")
+    elif expr.scalar_fn:
+      kind = expr.scalar_fn.kind.name.lower()
+      fn_name = expr.scalar_fn.fn_name
+      if expr.scalar_fn.attr_name:
+        fn_name, _ = self.fn_attr_mapping[expr.scalar_fn.attr_name]
+      fn = self._get_function(f"_{kind}_{fn_name}")
       operand_values = [
-          self.expression(operand) for operand in expr.scalar_apply.operands
+          self.expression(operand) for operand in expr.scalar_fn.operands
       ]
+      if expr.scalar_fn.kind == FunctionKind.TYPE:
+        operand_values = [expr.scalar_fn.type_var.name] + operand_values
       return fn(*operand_values)
-    elif expr.symbolic_cast:
-      operand_value = self.expression(expr.symbolic_cast.operand)
-      return self.cast(expr.symbolic_cast.to_type.name, operand_value,
-                       expr.symbolic_cast.is_unsigned_cast)
     raise NotImplementedError(f"Unimplemented scalar body expression: {expr}")
 
-  def cast(self, type_var_name: str, operand: Value,
-           is_unsigned_cast: bool) -> Value:
+  def yield_outputs(self, *output_names: str):
+    output_values = []
+    for n in output_names:
+      try:
+        output_values.append(self.yield_mapping[n])
+      except KeyError:
+        raise ValueError(f"Body assignments do not assign all outputs: "
+                         f"missing '{n}'")
+    linalg.YieldOp(output_values)
+
+  def _get_function(self, fn_name: str) -> Callable:
+    try:
+      fn = getattr(self, f"{fn_name}")
+    except AttributeError:
+      raise ValueError(f"Function '{fn_name}' is not a known function")
+    return fn
+
+  def _cast(self,
+            type_var_name: str,
+            operand: Value,
+            is_unsigned_cast: bool = False) -> Value:
     try:
       to_type = self.type_mapping[type_var_name]
     except KeyError:
@@ -291,74 +375,100 @@ class _BodyBuilder:
     raise ValueError(f"Unable to cast body expression from {operand_type} to "
                      f"{to_type}")
 
-  def yield_outputs(self, *output_names: str):
-    output_values = []
-    for n in output_names:
-      try:
-        output_values.append(self.yield_mapping[n])
-      except KeyError:
-        raise ValueError(f"Body assignments do not assign all outputs: "
-                         f"missing '{n}'")
-    linalg.YieldOp(output_values)
+  def _type_cast_signed(self, type_var_name: str, operand: Value) -> Value:
+    return self._cast(type_var_name, operand, False)
 
-  def _eval_add(self, lhs: Value, rhs: Value) -> Value:
-    if _is_floating_point_type(lhs.type):
-      return arith.AddFOp(lhs, rhs).result
-    if _is_integer_type(lhs.type) or _is_index_type(lhs.type):
-      return arith.AddIOp(lhs, rhs).result
-    raise NotImplementedError("Unsupported 'add' operand: {lhs}")
+  def _type_cast_unsigned(self, type_var_name: str, operand: Value) -> Value:
+    return self._cast(type_var_name, operand, True)
 
-  def _eval_exp(self, x: Value) -> Value:
+  def _unary_exp(self, x: Value) -> Value:
     if _is_floating_point_type(x.type):
       return math.ExpOp(x).result
     raise NotImplementedError("Unsupported 'exp' operand: {x}")
 
-  def _eval_log(self, x: Value) -> Value:
+  def _unary_log(self, x: Value) -> Value:
     if _is_floating_point_type(x.type):
       return math.LogOp(x).result
     raise NotImplementedError("Unsupported 'log' operand: {x}")
 
-  def _eval_sub(self, lhs: Value, rhs: Value) -> Value:
+  def _unary_abs(self, x: Value) -> Value:
+    if _is_floating_point_type(x.type):
+      return math.AbsFOp(x).result
+    raise NotImplementedError("Unsupported 'abs' operand: {x}")
+
+  def _unary_ceil(self, x: Value) -> Value:
+    if _is_floating_point_type(x.type):
+      return math.CeilOp(x).result
+    raise NotImplementedError("Unsupported 'ceil' operand: {x}")
+
+  def _unary_floor(self, x: Value) -> Value:
+    if _is_floating_point_type(x.type):
+      return math.FloorOp(x).result
+    raise NotImplementedError("Unsupported 'floor' operand: {x}")
+
+  def _unary_negf(self, x: Value) -> Value:
+    if _is_floating_point_type(x.type):
+      return arith.NegFOp(x).result
+    if _is_complex_type(x.type):
+      return complex.NegOp(x).result
+    raise NotImplementedError("Unsupported 'negf' operand: {x}")
+
+  def _binary_add(self, lhs: Value, rhs: Value) -> Value:
+    if _is_floating_point_type(lhs.type):
+      return arith.AddFOp(lhs, rhs).result
+    if _is_integer_type(lhs.type) or _is_index_type(lhs.type):
+      return arith.AddIOp(lhs, rhs).result
+    if _is_complex_type(lhs.type):
+      return complex.AddOp(lhs, rhs).result
+    raise NotImplementedError("Unsupported 'add' operands: {lhs}, {rhs}")
+
+  def _binary_sub(self, lhs: Value, rhs: Value) -> Value:
     if _is_floating_point_type(lhs.type):
       return arith.SubFOp(lhs, rhs).result
     if _is_integer_type(lhs.type) or _is_index_type(lhs.type):
       return arith.SubIOp(lhs, rhs).result
-    raise NotImplementedError("Unsupported 'sub' operand: {lhs}")
+    if _is_complex_type(lhs.type):
+      return complex.SubOp(lhs, rhs).result
+    raise NotImplementedError("Unsupported 'sub' operands: {lhs}, {rhs}")
 
-  def _eval_mul(self, lhs: Value, rhs: Value) -> Value:
+  def _binary_mul(self, lhs: Value, rhs: Value) -> Value:
     if _is_floating_point_type(lhs.type):
       return arith.MulFOp(lhs, rhs).result
     if _is_integer_type(lhs.type) or _is_index_type(lhs.type):
       return arith.MulIOp(lhs, rhs).result
-    raise NotImplementedError("Unsupported 'mul' operand: {lhs}")
+    if _is_complex_type(lhs.type):
+      return complex.MulOp(lhs, rhs).result
+    raise NotImplementedError("Unsupported 'mul' operands: {lhs}, {rhs}")
 
-  def _eval_max(self, lhs: Value, rhs: Value) -> Value:
+  def _binary_max_signed(self, lhs: Value, rhs: Value) -> Value:
     if _is_floating_point_type(lhs.type):
-      return std.MaxFOp(lhs, rhs).result
+      return arith.MaxFOp(lhs, rhs).result
     if _is_integer_type(lhs.type) or _is_index_type(lhs.type):
-      return std.MaxSIOp(lhs, rhs).result
-    raise NotImplementedError("Unsupported 'max' operand: {lhs}")
+      return arith.MaxSIOp(lhs, rhs).result
+    raise NotImplementedError("Unsupported 'max' operands: {lhs}, {rhs}")
 
-  def _eval_max_unsigned(self, lhs: Value, rhs: Value) -> Value:
+  def _binary_max_unsigned(self, lhs: Value, rhs: Value) -> Value:
     if _is_floating_point_type(lhs.type):
-      return std.MaxFOp(lhs, rhs).result
+      return arith.MaxFOp(lhs, rhs).result
     if _is_integer_type(lhs.type) or _is_index_type(lhs.type):
-      return std.MaxUIOp(lhs, rhs).result
-    raise NotImplementedError("Unsupported 'max_unsigned' operand: {lhs}")
+      return arith.MaxUIOp(lhs, rhs).result
+    raise NotImplementedError(
+        "Unsupported 'max_unsigned' operands: {lhs}, {rhs}")
 
-  def _eval_min(self, lhs: Value, rhs: Value) -> Value:
+  def _binary_min_signed(self, lhs: Value, rhs: Value) -> Value:
     if _is_floating_point_type(lhs.type):
-      return std.MinFOp(lhs, rhs).result
+      return arith.MinFOp(lhs, rhs).result
     if _is_integer_type(lhs.type) or _is_index_type(lhs.type):
-      return std.MinSIOp(lhs, rhs).result
-    raise NotImplementedError("Unsupported 'min' operand: {lhs}")
+      return arith.MinSIOp(lhs, rhs).result
+    raise NotImplementedError("Unsupported 'min' operands: {lhs}, {rhs}")
 
-  def _eval_min_unsigned(self, lhs: Value, rhs: Value) -> Value:
+  def _binary_min_unsigned(self, lhs: Value, rhs: Value) -> Value:
     if _is_floating_point_type(lhs.type):
-      return std.MinFOp(lhs, rhs).result
+      return arith.MinFOp(lhs, rhs).result
     if _is_integer_type(lhs.type) or _is_index_type(lhs.type):
-      return std.MinUIOp(lhs, rhs).result
-    raise NotImplementedError("Unsupported 'min_unsigned' operand: {lhs}")
+      return arith.MinUIOp(lhs, rhs).result
+    raise NotImplementedError(
+        "Unsupported 'min_unsigned' operands: {lhs}, {rhs}")
 
 
 def _infer_structured_outs(
@@ -411,6 +521,10 @@ def _add_type_mapping(operand_config: OperandDefConfig, operand_type: Type,
   block_arg_types.append(element_or_self_type)
 
 
+def _is_complex_type(t: Type) -> bool:
+  return ComplexType.isinstance(t)
+
+
 def _is_floating_point_type(t: Type) -> bool:
   # TODO: Create a FloatType in the Python API and implement the switch
   # there.
