@@ -224,6 +224,18 @@ extension SymbolGraph.Symbol {
         static let functionSignature = FunctionSignature.symbolCodingInfo
         static let spi = SPI.symbolCodingInfo
         static let snippet = Snippet.symbolCodingInfo
+        static let minimum = Minimum.symbolCodingInfo
+        static let maximum = Maximum.symbolCodingInfo
+        static let minimumExclusive = MinimumExclusive.symbolCodingInfo
+        static let maximumExclusive = MaximumExclusive.symbolCodingInfo
+        static let minimumLength = MinimumLength.symbolCodingInfo
+        static let maximumLength = MaximumLength.symbolCodingInfo
+        static let allowedValues = AllowedValues.symbolCodingInfo
+        static let defaultValue = DefaultValue.symbolCodingInfo
+        static let typeDetails = TypeDetails.symbolCodingInfo
+        static let httpEndpoint = HTTP.Endpoint.symbolCodingInfo
+        static let httpParameterSource = HTTP.ParameterSource.symbolCodingInfo
+        static let httpMediaType = HTTP.MediaType.symbolCodingInfo
         
         static let mixinCodingInfo: [String: SymbolMixinCodingInfo] = [
             CodingKeys.availability.codingKey.stringValue: Self.availability,
@@ -236,6 +248,18 @@ extension SymbolGraph.Symbol {
             CodingKeys.functionSignature.codingKey.stringValue: Self.functionSignature,
             CodingKeys.spi.codingKey.stringValue: Self.spi,
             CodingKeys.snippet.codingKey.stringValue: Self.snippet,
+            CodingKeys.minimum.codingKey.stringValue: Self.minimum,
+            CodingKeys.maximum.codingKey.stringValue: Self.maximum,
+            CodingKeys.minimumExclusive.codingKey.stringValue: Self.minimumExclusive,
+            CodingKeys.maximumExclusive.codingKey.stringValue: Self.maximumExclusive,
+            CodingKeys.minimumLength.codingKey.stringValue: Self.minimumLength,
+            CodingKeys.maximumLength.codingKey.stringValue: Self.maximumLength,
+            CodingKeys.allowedValues.codingKey.stringValue: Self.allowedValues,
+            CodingKeys.defaultValue.codingKey.stringValue: Self.defaultValue,
+            CodingKeys.typeDetails.codingKey.stringValue: Self.typeDetails,
+            CodingKeys.httpEndpoint.codingKey.stringValue: Self.httpEndpoint,
+            CodingKeys.httpParameterSource.codingKey.stringValue: Self.httpParameterSource,
+            CodingKeys.httpMediaType.codingKey.stringValue: Self.httpMediaType,
         ]
         
         static func == (lhs: SymbolGraph.Symbol.CodingKeys, rhs: SymbolGraph.Symbol.CodingKeys) -> Bool {
