@@ -1,4 +1,4 @@
-// swift-tools-version:5.5
+// swift-tools-version:5.7
 
 import PackageDescription
 import class Foundation.ProcessInfo
@@ -7,14 +7,14 @@ let macOSPlatform: SupportedPlatform
 if let deploymentTarget = ProcessInfo.processInfo.environment["SWIFTPM_MACOS_DEPLOYMENT_TARGET"] {
     macOSPlatform = .macOS(deploymentTarget)
 } else {
-    macOSPlatform = .macOS(.v10_15)
+    macOSPlatform = .macOS(.v12)
 }
 
 let package = Package(
   name: "swift-driver",
   platforms: [
     macOSPlatform,
-    .iOS(.v13),
+    .iOS(.v15),
   ],
   products: [
     .executable(
@@ -53,7 +53,7 @@ let package = Package(
         "SwiftOptions",
         .product(name: "SwiftToolsSupport-auto", package: "swift-tools-support-core"),
         "CSwiftScan",
-        .product(name: "Yams", package: "Yams"),
+        .product(name: "Yams", package: "yams"),
       ],
       exclude: ["CMakeLists.txt"]),
 
@@ -137,7 +137,7 @@ let package = Package(
 if ProcessInfo.processInfo.environment["SWIFT_DRIVER_LLBUILD_FWK"] == nil {
     if ProcessInfo.processInfo.environment["SWIFTCI_USE_LOCAL_DEPS"] == nil {
         package.dependencies += [
-            .package(url: "https://github.com/apple/swift-llbuild.git", .branch("release/5.9")),
+            .package(url: "https://github.com/apple/swift-llbuild.git", branch: "release/5.10"),
         ]
         package.targets.first(where: { $0.name == "SwiftDriverExecution" })!.dependencies += [
             .product(name: "llbuildSwift", package: "swift-llbuild"),
@@ -156,7 +156,7 @@ if ProcessInfo.processInfo.environment["SWIFT_DRIVER_LLBUILD_FWK"] == nil {
 
 if ProcessInfo.processInfo.environment["SWIFTCI_USE_LOCAL_DEPS"] == nil {
   package.dependencies += [
-    .package(url: "https://github.com/apple/swift-tools-support-core.git", .branch("release/5.9")),
+    .package(url: "https://github.com/apple/swift-tools-support-core.git", branch: "release/5.10"),
     .package(url: "https://github.com/jpsim/Yams.git", .upToNextMinor(from: "5.0.0")),
     // The 'swift-argument-parser' version declared here must match that
     // used by 'swift-package-manager' and 'sourcekit-lsp'. Please coordinate
@@ -166,7 +166,7 @@ if ProcessInfo.processInfo.environment["SWIFTCI_USE_LOCAL_DEPS"] == nil {
 } else {
     package.dependencies += [
         .package(path: "../swift-tools-support-core"),
-        .package(name: "Yams", path: "../yams"),
+        .package(path: "../yams"),
         .package(path: "../swift-argument-parser"),
     ]
 }
