@@ -18,7 +18,7 @@
 #include <stdint.h>
 
 #define SWIFTSCAN_VERSION_MAJOR 0
-#define SWIFTSCAN_VERSION_MINOR 1
+#define SWIFTSCAN_VERSION_MINOR 5
 
 //=== Public Scanner Data Types -------------------------------------------===//
 
@@ -77,6 +77,19 @@ typedef struct {
 typedef struct swiftscan_scan_invocation_s *swiftscan_scan_invocation_t;
 typedef void *swiftscan_scanner_t;
 
+//=== CAS/Caching Specification -------------------------------------------===//
+typedef struct swiftscan_cas_s *swiftscan_cas_t;
+typedef struct swiftscan_cas_options_s *swiftscan_cas_options_t;
+
+typedef enum {
+  SWIFTSCAN_OUTPUT_TYPE_OBJECT = 0,
+  SWIFTSCAN_OUTPUT_TYPE_SWIFTMODULE = 1,
+  SWIFTSCAN_OUTPUT_TYPE_SWIFTINTERFACE = 2,
+  SWIFTSCAN_OUTPUT_TYPE_SWIFTPRIVATEINTERFACE = 3,
+  SWIFTSCAN_OUTPUT_TYPE_CLANG_MODULE = 4,
+  SWIFTSCAN_OUTPUT_TYPE_CLANG_PCH = 5
+} swiftscan_output_kind_t;
+
 //=== libSwiftScan Functions ------------------------------------------------===//
 
 typedef struct {
@@ -98,7 +111,7 @@ typedef struct {
   (*swiftscan_module_info_get_direct_dependencies)(swiftscan_dependency_info_t);
   swiftscan_module_details_t
   (*swiftscan_module_info_get_details)(swiftscan_dependency_info_t);
-  
+
   //=== Dependency Module Info Details Functions ----------------------------===//
   swiftscan_dependency_info_kind_t
   (*swiftscan_module_detail_get_kind)(swiftscan_module_details_t);
@@ -117,6 +130,8 @@ typedef struct {
   swiftscan_string_set_t *
   (*swiftscan_swift_textual_detail_get_command_line)(swiftscan_module_details_t);
   swiftscan_string_set_t *
+  (*swiftscan_swift_textual_detail_get_bridging_pch_command_line)(swiftscan_module_details_t);
+  swiftscan_string_set_t *
   (*swiftscan_swift_textual_detail_get_extra_pcm_args)(swiftscan_module_details_t);
   swiftscan_string_ref_t
   (*swiftscan_swift_textual_detail_get_context_hash)(swiftscan_module_details_t);
@@ -124,6 +139,8 @@ typedef struct {
   (*swiftscan_swift_textual_detail_get_is_framework)(swiftscan_module_details_t);
   swiftscan_string_set_t *
   (*swiftscan_swift_textual_detail_get_swift_overlay_dependencies)(swiftscan_module_details_t);
+  swiftscan_string_ref_t
+  (*swiftscan_swift_textual_detail_get_module_cache_key)(swiftscan_module_details_t);
 
   //=== Swift Binary Module Details query APIs ------------------------------===//
   swiftscan_string_ref_t
@@ -136,6 +153,8 @@ typedef struct {
   (*swiftscan_swift_binary_detail_get_header_dependencies)(swiftscan_module_details_t);
   bool
   (*swiftscan_swift_binary_detail_get_is_framework)(swiftscan_module_details_t);
+  swiftscan_string_ref_t
+  (*swiftscan_swift_binary_detail_get_module_cache_key)(swiftscan_module_details_t);
 
   //=== Swift Placeholder Module Details query APIs -------------------------===//
   swiftscan_string_ref_t
@@ -154,6 +173,8 @@ typedef struct {
   (*swiftscan_clang_detail_get_command_line)(swiftscan_module_details_t);
   swiftscan_string_set_t *
   (*swiftscan_clang_detail_get_captured_pcm_args)(swiftscan_module_details_t);
+  swiftscan_string_ref_t
+  (*swiftscan_clang_detail_get_module_cache_key)(swiftscan_module_details_t);
 
   //=== Batch Scan Input Functions ------------------------------------------===//
   swiftscan_batch_scan_input_t *
@@ -224,7 +245,7 @@ typedef struct {
   swiftscan_string_set_t *
   (*swiftscan_compiler_supported_features_query)(void);
 
-  //=== Scanner Functions ---------------------------------------------------===//  
+  //=== Scanner Functions ---------------------------------------------------===//
   swiftscan_scanner_t (*swiftscan_scanner_create)(void);
   void (*swiftscan_scanner_dispose)(swiftscan_scanner_t);
   swiftscan_dependency_graph_t
@@ -235,7 +256,7 @@ typedef struct {
                                         swiftscan_scan_invocation_t);
   swiftscan_import_set_t
   (*swiftscan_import_set_create)(swiftscan_scanner_t, swiftscan_scan_invocation_t);
-  
+
   //=== Scanner Diagnostics -------------------------------------------------===//
   swiftscan_diagnostic_set_t*
   (*swiftscan_scanner_diagnostics_query)(swiftscan_scanner_t);
@@ -253,6 +274,26 @@ typedef struct {
   bool (*swiftscan_scanner_cache_load)(swiftscan_scanner_t scanner, const char * path);
   void (*swiftscan_scanner_cache_reset)(swiftscan_scanner_t scanner);
 
+  //=== Scanner CAS Operations ----------------------------------------------===//
+  swiftscan_cas_options_t (*swiftscan_cas_options_create)(void);
+  void (*swiftscan_cas_options_dispose)(swiftscan_cas_options_t options);
+  void (*swiftscan_cas_options_set_ondisk_path)(swiftscan_cas_options_t options,
+                                                const char *path);
+  void (*swiftscan_cas_options_set_plugin_path)(swiftscan_cas_options_t options,
+                                                const char *path);
+  bool (*swiftscan_cas_options_set_option)(swiftscan_cas_options_t options,
+                                           const char *name, const char *value,
+                                           swiftscan_string_ref_t *error);
+  swiftscan_cas_t (*swiftscan_cas_create_from_options)(
+      swiftscan_cas_options_t options, swiftscan_string_ref_t *error);
+  void (*swiftscan_cas_dispose)(swiftscan_cas_t cas);
+  swiftscan_string_ref_t (*swiftscan_cas_store)(swiftscan_cas_t cas,
+                                                uint8_t *data, unsigned size,
+                                                swiftscan_string_ref_t *error);
+  swiftscan_string_ref_t (*swiftscan_compute_cache_key)(
+      swiftscan_cas_t cas, int argc, const char *argv, const char *input,
+      swiftscan_output_kind_t, swiftscan_string_ref_t *error);
+
 } swiftscan_functions_t;
 
 #endif // SWIFT_C_DEPENDENCY_SCAN_H
