@@ -54,6 +54,21 @@ import func TSCBasic.topologicalSort
 }
 
 extension InterModuleDependencyGraph {
+  var topologicalSorting: [ModuleDependencyId] {
+    get throws {
+      try topologicalSort(Array(modules.keys),
+                          successors: {
+        var dependencies: [ModuleDependencyId] = []
+        let moduleInfo = try moduleInfo(of: $0)
+        dependencies.append(contentsOf: moduleInfo.directDependencies ?? [])
+        if case .swift(let swiftModuleDetails) = moduleInfo.details {
+          dependencies.append(contentsOf: swiftModuleDetails.swiftOverlayDependencies ?? [])
+        }
+        return dependencies
+      })
+    }
+  }
+
   /// Compute a set of modules that are "reachable" (form direct or transitive dependency)
   /// from each module in the graph.
   /// This routine relies on the fact that the dependency graph is acyclic. A lack of cycles means
@@ -64,10 +79,8 @@ extension InterModuleDependencyGraph {
   ///     T(v) = T(v) ∪ T(w)
   ///   }
   /// }
-  func computeTransitiveClosure() throws -> [ModuleDependencyId : Set<ModuleDependencyId>] {
-    let topologicalIdList =
-      try topologicalSort(Array(modules.keys),
-                          successors: { try moduleInfo(of: $0).directDependencies! })
+  @_spi(Testing) public func computeTransitiveClosure() throws -> [ModuleDependencyId : Set<ModuleDependencyId>] {
+    let topologicalIdList = try self.topologicalSorting
     // This structure will contain the final result
     var transitiveClosureMap =
       topologicalIdList.reduce(into: [ModuleDependencyId : Set<ModuleDependencyId>]()) {
@@ -267,12 +280,13 @@ internal extension InterModuleDependencyGraph {
       return
     }
 
-    // If no more dependencies, this is a leaf node, we are done
-    guard let dependencies = sourceInfo.directDependencies else {
-      return
+    var allDependencies = sourceInfo.directDependencies ?? []
+    if case .swift(let swiftModuleDetails) = sourceInfo.details,
+          let overlayDependencies = swiftModuleDetails.swiftOverlayDependencies {
+      allDependencies.append(contentsOf: overlayDependencies)
     }
 
-    for dependency in dependencies {
+    for dependency in allDependencies {
       try findAllPaths(source: dependency,
                        to: moduleName,
                        pathSoFar: pathSoFar + [dependency],
