@@ -30,6 +30,15 @@ public enum ModuleDependencyId: Hashable {
     case .clang(let name): return name
     }
   }
+
+  internal var moduleNameForDiagnostic: String {
+    switch self {
+    case .swift(let name): return name
+    case .swiftPlaceholder(let name): return name + "(placeholder)"
+    case .swiftPrebuiltExternal(let name): return name + "(swiftmodule)"
+    case .clang(let name): return name + "(pcm)"
+    }
+  }
 }
 
 extension ModuleDependencyId: Codable {
@@ -108,6 +117,9 @@ public struct SwiftModuleDetails: Codable {
   /// Options to the compile command
   public var commandLine: [String]? = []
 
+  /// Options to the compile command
+  public var bridgingPchCommandLine: [String]? = []
+
   /// The context hash for this module that encodes the producing interface's path,
   /// target triple, etc. This field is optional because it is absent for the ModuleInfo
   /// corresponding to the main module being built.
@@ -123,6 +135,9 @@ public struct SwiftModuleDetails: Codable {
 
   /// A set of Swift Overlays of Clang Module Dependencies
   var swiftOverlayDependencies: [ModuleDependencyId]?
+
+  /// The module cache key of the output module.
+  public var moduleCacheKey: String?
 }
 
 /// Details specific to Swift placeholder dependencies.
@@ -152,16 +167,20 @@ public struct SwiftPrebuiltExternalModuleDetails: Codable {
   /// A flag to indicate whether or not this module is a framework.
   public var isFramework: Bool?
 
+  /// The module cache key of the pre-built module.
+  public var moduleCacheKey: String?
+
   public init(compiledModulePath: TextualVirtualPath,
               moduleDocPath: TextualVirtualPath? = nil,
               moduleSourceInfoPath: TextualVirtualPath? = nil,
               headerDependencies: [TextualVirtualPath]? = nil,
-              isFramework: Bool) throws {
+              isFramework: Bool, moduleCacheKey: String? = nil) throws {
     self.compiledModulePath = compiledModulePath
     self.moduleDocPath = moduleDocPath
     self.moduleSourceInfoPath = moduleSourceInfoPath
     self.headerDependencyPaths = headerDependencies
     self.isFramework = isFramework
+    self.moduleCacheKey = moduleCacheKey
   }
 }
 
@@ -180,14 +199,19 @@ public struct ClangModuleDetails: Codable {
   /// are covered by the directDependencies info of this module
   public var capturedPCMArgs: Set<[String]>?
 
+  /// The module cache key of the output module.
+  public var moduleCacheKey: String?
+
   public init(moduleMapPath: TextualVirtualPath,
               contextHash: String,
               commandLine: [String],
-              capturedPCMArgs: Set<[String]>?) {
+              capturedPCMArgs: Set<[String]>?,
+              moduleCacheKey: String? = nil) {
     self.moduleMapPath = moduleMapPath
     self.contextHash = contextHash
     self.commandLine = commandLine
     self.capturedPCMArgs = capturedPCMArgs
+    self.moduleCacheKey = moduleCacheKey
   }
 }
 
@@ -311,7 +335,7 @@ internal extension InterModuleDependencyGraph {
 
 public struct InterModuleDependencyImports: Codable {
   public var imports: [String]
-  
+
   public init(imports: [String], moduleAliases: [String: String]? = nil) {
     var realImports = [String]()
     if let aliases = moduleAliases {
