@@ -106,7 +106,7 @@ public extension Driver {
     var commandLine: [Job.ArgTemplate] = swiftCompilerPrefixArgs.map { Job.ArgTemplate.flag($0) }
     commandLine.appendFlag("-frontend")
     commandLine.appendFlag("-scan-dependencies")
-    try addCommonFrontendOptions(commandLine: &commandLine, inputs: &inputs,
+    try addCommonFrontendOptions(commandLine: &commandLine, inputs: &inputs, kind: .scanDependencies,
                                  bridgingHeaderHandling: .parsed,
                                  moduleDependencyGraphUse: .dependencyScan)
     // FIXME: MSVC runtime flags
@@ -120,6 +120,11 @@ public extension Driver {
       commandLine.appendPath(dependencyPlaceholderMapFile)
     }
 
+    try commandLine.appendLast(.clangIncludeTree, from: &parsedOptions)
+    if isFrontendArgSupported(.clangScannerModuleCachePath) {
+      try commandLine.appendLast(.clangScannerModuleCachePath, from: &parsedOptions)
+    }
+
     // Pass on the input files
     commandLine.append(contentsOf: inputFiles.filter { $0.type == .swift }.map { .path($0.file) })
     return (inputs, commandLine)
@@ -144,7 +149,7 @@ public extension Driver {
   }
 
   /// Returns false if the lib is available and ready to use
-  private func initSwiftScanLib() throws -> Bool {
+  private mutating func initSwiftScanLib() throws -> Bool {
     // If `-nonlib-dependency-scanner` was specified or the libSwiftScan library cannot be found,
     // attempt to fallback to using `swift-frontend -scan-dependencies` invocations for dependency
     // scanning.
@@ -161,6 +166,11 @@ public extension Driver {
         diagnosticEngine.emit(.warn_scanner_frontend_fallback())
       }
     }
+    if !fallbackToFrontend && enableCaching {
+      try interModuleDependencyOracle.createCAS(pluginPath: try getCASPluginPath(),
+                                                onDiskPath: try getOnDiskCASPath(),
+                                                pluginOptions: try getCASPluginOptions())
+    }
     return fallbackToFrontend
   }
 
@@ -210,7 +220,7 @@ public extension Driver {
     let scannerJob = try dependencyScanningJob()
     let forceResponseFiles = parsedOptions.hasArgument(.driverForceResponseFiles)
     let dependencyGraph: InterModuleDependencyGraph
-    
+
     if parsedOptions.contains(.v) {
       let arguments: [String] = try executor.resolver.resolveArgumentList(for: scannerJob,
                                                                           useResponseFiles: .disabled)
@@ -336,7 +346,7 @@ public extension Driver {
     commandLine.appendFlag("-frontend")
     commandLine.appendFlag("-scan-dependencies")
     commandLine.appendFlag("-import-prescan")
-    try addCommonFrontendOptions(commandLine: &commandLine, inputs: &inputs,
+    try addCommonFrontendOptions(commandLine: &commandLine, inputs: &inputs, kind: .scanDependencies,
                                  bridgingHeaderHandling: .parsed,
                                  moduleDependencyGraphUse: .dependencyScan)
     // FIXME: MSVC runtime flags
@@ -366,7 +376,7 @@ public extension Driver {
     // is present.
     commandLine.appendFlag("-frontend")
     commandLine.appendFlag("-scan-dependencies")
-    try addCommonFrontendOptions(commandLine: &commandLine, inputs: &inputs,
+    try addCommonFrontendOptions(commandLine: &commandLine, inputs: &inputs, kind: .scanDependencies,
                                  bridgingHeaderHandling: .ignored,
                                  moduleDependencyGraphUse: .dependencyScan)
 
