@@ -28,16 +28,19 @@
   public let prebuiltHeaderDependencyPaths: [TextualVirtualPath]?
   /// A flag to indicate whether this module is a framework
   public let isFramework: Bool
+  /// The cache key for the module.
+  public let moduleCacheKey: String?
 
   init(name: String, modulePath: TextualVirtualPath, docPath: TextualVirtualPath? = nil,
        sourceInfoPath: TextualVirtualPath? = nil, headerDependencies: [TextualVirtualPath]? = nil,
-       isFramework: Bool = false) {
+       isFramework: Bool = false, moduleCacheKey: String? = nil) {
     self.moduleName = name
     self.modulePath = modulePath
     self.docPath = docPath
     self.sourceInfoPath = sourceInfoPath
     self.prebuiltHeaderDependencyPaths = headerDependencies
     self.isFramework = isFramework
+    self.moduleCacheKey = moduleCacheKey
   }
 }
 
@@ -54,12 +57,16 @@
   public let clangModuleMapPath: TextualVirtualPath
   /// A flag to indicate whether this module is a framework
   public let isFramework: Bool
+  /// The cache key for the module.
+  public let clangModuleCacheKey: String?
 
-  init(name: String, modulePath: TextualVirtualPath, moduleMapPath: TextualVirtualPath) {
+  init(name: String, modulePath: TextualVirtualPath, moduleMapPath: TextualVirtualPath,
+       moduleCacheKey: String? = nil) {
     self.moduleName = name
     self.clangModulePath = modulePath
     self.clangModuleMapPath = moduleMapPath
     self.isFramework = false
+    self.clangModuleCacheKey = moduleCacheKey
   }
 }
 
@@ -120,3 +127,15 @@ public extension BatchScanModuleInfo {
     }
   }
 }
+
+extension SwiftModuleArtifactInfo: Comparable {
+  public static func < (lhs: SwiftModuleArtifactInfo, rhs: SwiftModuleArtifactInfo) -> Bool {
+    return lhs.moduleName < rhs.moduleName
+  }
+}
+
+extension ClangModuleArtifactInfo: Comparable {
+  public static func < (lhs: ClangModuleArtifactInfo, rhs: ClangModuleArtifactInfo) -> Bool {
+    return lhs.moduleName < rhs.moduleName
+  }
+}
