@@ -225,6 +225,7 @@ extension Driver {
   mutating func compileJob(primaryInputs: [TypedVirtualPath],
                            outputType: FileType?,
                            addJobOutputs: ([TypedVirtualPath]) -> Void,
+                           pchCompileJob: Job?,
                            emitModuleTrace: Bool)
   throws -> Job {
     var commandLine: [Job.ArgTemplate] = swiftCompilerPrefixArgs.map { Job.ArgTemplate.flag($0) }
@@ -283,7 +284,8 @@ extension Driver {
       commandLine.appendFlag(.disableObjcAttrRequiresFoundationModule)
     }
 
-    try addCommonFrontendOptions(commandLine: &commandLine, inputs: &inputs)
+    try addCommonFrontendOptions(commandLine: &commandLine, inputs: &inputs, kind: .compile)
+
     // FIXME: MSVC runtime flags
 
     if Driver.canDoCrossModuleOptimization(parsedOptions: &parsedOptions) &&
@@ -370,13 +372,8 @@ extension Driver {
     try commandLine.appendLast(.disableAutolinkingRuntimeCompatibilityConcurrency, from: &parsedOptions)
     try commandLine.appendLast(.checkApiAvailabilityOnly, from: &parsedOptions)
 
-    if compilerMode.isSingleCompilation {
-      try commandLine.appendLast(.emitSymbolGraph, from: &parsedOptions)
-      try commandLine.appendLast(.emitSymbolGraphDir, from: &parsedOptions)
-    }
-    try commandLine.appendLast(.includeSpiSymbols, from: &parsedOptions)
-    try commandLine.appendLast(.emitExtensionBlockSymbols, .omitExtensionBlockSymbols, from: &parsedOptions)
-    try commandLine.appendLast(.symbolGraphMinimumAccessLevel, from: &parsedOptions)
+    try addCommonSymbolGraphOptions(commandLine: &commandLine,
+                                    includeGraph: compilerMode.isSingleCompilation)
 
     addJobOutputs(outputs)
 
@@ -385,6 +382,7 @@ extension Driver {
       let pchInput = TypedVirtualPath(file: pchPath, type: .pch)
       inputs.append(pchInput)
     }
+    try addBridgingHeaderPCHCacheKeyArguments(commandLine: &commandLine, pchCompileJob: pchCompileJob)
 
     let displayInputs : [TypedVirtualPath]
     if case .singleCompile = compilerMode {
