@@ -15,6 +15,7 @@ import SwiftOptions
 import class TSCBasic.LocalFileOutputByteStream
 import class TSCBasic.TerminalController
 import struct TSCBasic.RelativePath
+import struct TSCBasic.AbsolutePath
 import var TSCBasic.stderrStream
 
 /// Whether we should produce color diagnostics by default.
@@ -54,9 +55,9 @@ extension Driver {
   mutating func addCommonFrontendOptions(
     commandLine: inout [Job.ArgTemplate],
     inputs: inout [TypedVirtualPath],
+    kind: Job.Kind,
     bridgingHeaderHandling: BridgingHeaderHandling = .precompiled,
-    moduleDependencyGraphUse: ModuleDependencyGraphUse = .computed,
-    isGeneratePCH: Bool = false
+    moduleDependencyGraphUse: ModuleDependencyGraphUse = .computed
   ) throws {
     // Only pass -target to the REPL or immediate modes if it was explicitly
     // specified on the command line.
@@ -79,10 +80,19 @@ extension Driver {
     // May also be used for generation of the dependency graph itself in ExplicitModuleBuild mode.
     if (parsedOptions.contains(.driverExplicitModuleBuild) &&
           moduleDependencyGraphUse == .computed) {
-      if isGeneratePCH {
+      switch kind {
+      case .generatePCH:
         try addExplicitPCHBuildArguments(inputs: &inputs, commandLine: &commandLine)
-      } else {
+      case .compile, .emitModule, .interpret, .verifyModuleInterface:
         try addExplicitModuleBuildArguments(inputs: &inputs, commandLine: &commandLine)
+      case .backend, .mergeModule, .compileModuleFromInterface,
+           .generatePCM, .dumpPCM, .repl, .printTargetInfo,
+           .versionRequest, .autolinkExtract, .generateDSYM,
+           .help, .link, .verifyDebugInfo, .scanDependencies,
+           .emitSupportedFeatures, .moduleWrap,
+           .generateAPIBaseline, .generateABIBaseline, .compareAPIBaseline,
+           .compareABIBaseline:
+        break // Do not support creating from dependency scanner output.
       }
     }
 
@@ -97,15 +107,6 @@ extension Driver {
       commandLine.appendFlag("-aarch64-use-tbi")
     }
 
-    // Potentially unavailable enum cases are downgraded to a warning when building a
-    // swiftmodule, to allow building a module (or module interface) for an older
-    // deployment target than the framework itself.
-    if isFrontendArgSupported(.warnOnPotentiallyUnavailableEnumCase) {
-      if compilerOutputType == .swiftModule {
-        commandLine.appendFlag(.warnOnPotentiallyUnavailableEnumCase)
-      }
-    }
-
     // Enable or disable ObjC interop appropriately for the platform
     if targetTriple.isDarwin {
       commandLine.appendFlag(.enableObjcInterop)
@@ -150,6 +151,11 @@ extension Driver {
     try commandLine.appendAll(.F, .Fsystem, from: &parsedOptions)
     try commandLine.appendAll(.vfsoverlay, from: &parsedOptions)
 
+    if let gccToolchain = parsedOptions.getLastArgument(.gccToolchain) {
+        commandLine.appendFlag(.Xcc)
+        commandLine.appendFlag("--gcc-toolchain=\(gccToolchain.asSingle)")
+    }
+
     try commandLine.appendLast(.AssertConfig, from: &parsedOptions)
     try commandLine.appendLast(.autolinkForceLoad, from: &parsedOptions)
 
@@ -289,7 +295,11 @@ extension Driver {
       try commandLine.appendLast(.enableBuiltinModule, from: &parsedOptions)
     }
 
-    if let workingDirectory = workingDirectory {
+    if isFrontendArgSupported(.disableSandbox) {
+      try commandLine.appendLast(.disableSandbox, from: &parsedOptions)
+    }
+
+    if !useClangIncludeTree, let workingDirectory = workingDirectory {
       // Add -Xcc -working-directory before any other -Xcc options to ensure it is
       // overridden by an explicit -Xcc -working-directory, although having a
       // different working directory is probably incorrect.
@@ -327,9 +337,30 @@ extension Driver {
       try commandLine.appendAll(.fileCompilationDir, from: &parsedOptions)
     }
 
+    // CAS related options.
+    if enableCaching {
+      commandLine.appendFlag(.cacheCompileJob)
+      if let casPath = try getOnDiskCASPath() {
+        commandLine.appendFlag(.casPath)
+        commandLine.appendFlag(casPath.pathString)
+      }
+      if let pluginPath = try getCASPluginPath() {
+        commandLine.appendFlag(.casPluginPath)
+        commandLine.appendFlag(pluginPath.pathString)
+      }
+      try commandLine.appendAll(.casPluginOption, from: &parsedOptions)
+      try commandLine.appendLast(.cacheRemarks, from: &parsedOptions)
+    }
+    if useClangIncludeTree {
+      commandLine.appendFlag(.clangIncludeTree)
+    }
+
     // Pass through any subsystem flags.
     try commandLine.appendAll(.Xllvm, from: &parsedOptions)
-    try commandLine.appendAll(.Xcc, from: &parsedOptions)
+
+    if !useClangIncludeTree {
+      try commandLine.appendAll(.Xcc, from: &parsedOptions)
+    }
 
     if let importedObjCHeader = importedObjCHeader,
         bridgingHeaderHandling != .ignored {
@@ -391,6 +422,20 @@ extension Driver {
     }
   }
 
+  func addBridgingHeaderPCHCacheKeyArguments(commandLine: inout [Job.ArgTemplate],
+                                             pchCompileJob: Job?) throws {
+    guard let pchJob = pchCompileJob, enableCaching else { return }
+
+    // The pch input file (the bridging header) is added as last inputs to the job.
+    guard let inputFile = pchJob.inputs.last else { assertionFailure("no input files from pch job"); return }
+    assert(inputFile.type == .objcHeader, "Expect objc header input type")
+    let bridgingHeaderCacheKey = try interModuleDependencyOracle.computeCacheKeyForOutput(kind: .pch,
+                                                                                          commandLine: pchJob.commandLine,
+                                                                                          input: inputFile.fileHandle)
+    commandLine.appendFlag("-bridging-header-pch-key")
+    commandLine.appendFlag(bridgingHeaderCacheKey)
+  }
+
   mutating func addFrontendSupplementaryOutputArguments(commandLine: inout [Job.ArgTemplate],
                                                         primaryInputs: [TypedVirtualPath],
                                                         inputsGeneratingCodeCount: Int,
@@ -610,6 +655,17 @@ extension Driver {
     return flaggedInputOutputPairs.map { $0.output }
   }
 
+  mutating func addCommonSymbolGraphOptions(commandLine: inout [Job.ArgTemplate],
+                                            includeGraph: Bool = true) throws {
+    if includeGraph {
+      try commandLine.appendLast(.emitSymbolGraph, from: &parsedOptions)
+      try commandLine.appendLast(.emitSymbolGraphDir, from: &parsedOptions)
+    }
+    try commandLine.appendLast(.includeSpiSymbols, from: &parsedOptions)
+    try commandLine.appendLast(.emitExtensionBlockSymbols, .omitExtensionBlockSymbols, from: &parsedOptions)
+    try commandLine.appendLast(.symbolGraphMinimumAccessLevel, from: &parsedOptions)
+  }
+
   func addEntry(_ entries: inout [VirtualPath.Handle: [FileType: VirtualPath.Handle]], input: TypedVirtualPath?, output: TypedVirtualPath) {
     let entryInput: VirtualPath.Handle
     if let input = input?.fileHandle, input != OutputFileMap.singleInputKey {
@@ -640,6 +696,14 @@ extension Driver {
     try dependencyPlanner.resolveBridgingHeaderDependencies(inputs: &inputs, commandLine: &commandLine)
   }
 
+  /// If explicit dependency planner supports creating bridging header pch command.
+  public func supportsBridgingHeaderPCHCommand() throws -> Bool {
+    guard let dependencyPlanner = explicitDependencyBuildPlanner else {
+      return false
+    }
+    return try dependencyPlanner.supportsBridgingHeaderPCHCommand()
+  }
+
   /// In Explicit Module Build mode, distinguish between main module jobs and intermediate dependency build jobs,
   /// such as Swift modules built from .swiftmodule files and Clang PCMs.
   public func isExplicitMainModuleJob(job: Job) -> Bool {
