@@ -13,18 +13,28 @@
 import struct TSCBasic.RelativePath
 
 extension Driver {
+  mutating func addGeneratePCHFlags(commandLine: inout [Job.ArgTemplate], inputs: inout [TypedVirtualPath]) throws {
+    commandLine.appendFlag("-frontend")
+
+    try addCommonFrontendOptions(
+      commandLine: &commandLine, inputs: &inputs, kind: .generatePCH, bridgingHeaderHandling: .parsed)
+
+    try commandLine.appendLast(.indexStorePath, from: &parsedOptions)
+
+    commandLine.appendFlag(.emitPch)
+  }
+
   mutating func generatePCHJob(input: TypedVirtualPath, output: TypedVirtualPath) throws -> Job {
     var inputs = [TypedVirtualPath]()
     var outputs = [TypedVirtualPath]()
 
     var commandLine: [Job.ArgTemplate] = swiftCompilerPrefixArgs.map { Job.ArgTemplate.flag($0) }
 
-    commandLine.appendFlag("-frontend")
-
-    try addCommonFrontendOptions(
-      commandLine: &commandLine, inputs: &inputs, bridgingHeaderHandling: .parsed, isGeneratePCH: true)
-
-    try commandLine.appendLast(.indexStorePath, from: &parsedOptions)
+    if try supportsBridgingHeaderPCHCommand() {
+      try addExplicitPCHBuildArguments(inputs: &inputs, commandLine: &commandLine)
+    } else {
+      try addGeneratePCHFlags(commandLine: &commandLine, inputs: &inputs)
+    }
 
     // TODO: Should this just be pch output with extension changed?
     if parsedOptions.hasArgument(.serializeDiagnostics), let outputDirectory = parsedOptions.getLastArgument(.pchOutputDir)?.asSingle {
@@ -46,13 +56,7 @@ extension Driver {
       outputs.append(.init(file: path.intern(), type: .diagnostics))
     }
 
-    inputs.append(input)
-    commandLine.appendPath(input.file)
-
-    try commandLine.appendLast(.indexStorePath, from: &parsedOptions)
-
-    commandLine.appendFlag(.emitPch)
-
+    // New compute and add inputs and outputs.
     if parsedOptions.hasArgument(.pchOutputDir) &&
        !parsedOptions.contains(.driverExplicitModuleBuild) {
       try commandLine.appendLast(.pchOutputDir, from: &parsedOptions)
@@ -62,6 +66,9 @@ extension Driver {
     }
     outputs.append(output)
 
+    inputs.append(input)
+    commandLine.appendPath(input.file)
+
     return Job(
       moduleName: moduleOutputInfo.name,
       kind: .generatePCH,
