@@ -26,9 +26,14 @@ extension Toolchain {
     for targetInfo: FrontendTargetInfo,
     parsedOptions: inout ParsedOptions
   ) throws -> VirtualPath {
+    var platform = targetInfo.target.triple.platformName(conflatingDarwin: true)!
+    // compiler-rt moved these Android sanitizers into `lib/linux/` a couple
+    // years ago, llvm/llvm-project@a68ccba, so look for them there instead.
+    if platform == "android" {
+      platform = "linux"
+    }
     return VirtualPath.lookup(targetInfo.runtimeResourcePath.path)
-      .appending(components: "clang", "lib",
-                 targetInfo.target.triple.platformName(conflatingDarwin: true)!)
+      .appending(components: "clang", "lib", platform)
   }
 
   func runtimeLibraryPaths(
@@ -101,11 +106,13 @@ extension DarwinToolchain {
 
     // Link compatibility libraries, if we're deploying back to OSes that
     // have an older Swift runtime.
-    func addArgsForBackDeployLib(_ libName: String) throws {
+    func addArgsForBackDeployLib(_ libName: String, forceLoad: Bool) throws {
       let backDeployLibPath = VirtualPath.lookup(targetInfo.runtimeResourcePath.path)
         .appending(components: targetTriple.platformName() ?? "", libName)
       if try fileSystem.exists(backDeployLibPath) {
-        commandLine.append(.flag("-force_load"))
+        if forceLoad {
+          commandLine.append(.flag("-force_load"))
+        }
         commandLine.appendPath(backDeployLibPath)
       }
     }
@@ -122,7 +129,9 @@ extension DarwinToolchain {
       }
 
       if shouldLink {
-        try addArgsForBackDeployLib("lib" + compatibilityLib.libraryName + ".a")
+        // Old frontends don't set forceLoad at all; assume it's true in that case
+        try addArgsForBackDeployLib("lib" + compatibilityLib.libraryName + ".a",
+                                    forceLoad: compatibilityLib.forceLoad ?? true)
       }
     }
 
