@@ -43,6 +43,10 @@ extension WebAssemblyToolchain {
         commandLine.appendFlag("-fuse-ld=\(linkerArg)")
       }
 
+      if let arg = parsedOptions.getLastArgument(.ldPath)?.asSingle {
+        commandLine.append(.joinedOptionAndPath("--ld-path=", try VirtualPath(path: arg)))
+      }
+
       // Configure the toolchain.
       //
       // By default use the system `clang` to perform the link.  We use `clang` for
