@@ -105,6 +105,10 @@ extension WindowsToolchain {
       commandLine.appendFlag("-fuse-ld=lld")
     }
 
+    if let arg = parsedOptions.getLastArgument(.ldPath)?.asSingle {
+      commandLine.append(.joinedOptionAndPath("--ld-path=", try VirtualPath(path: arg)))
+    }
+
     switch lto {
     case .some(.llvmThin):
       commandLine.appendFlag("-flto=thin")
@@ -128,8 +132,10 @@ extension WindowsToolchain {
 
     // Since Windows has separate libraries per architecture, link against the
     // architecture specific version of the static library.
-    commandLine.appendFlag(.L)
-    commandLine.appendPath(VirtualPath.lookup(targetInfo.runtimeLibraryImportPaths.last!.path))
+    for libpath in targetInfo.runtimeLibraryImportPaths {
+      commandLine.appendFlag(.L)
+      commandLine.appendPath(VirtualPath.lookup(libpath.path))
+    }
 
     if !parsedOptions.hasArgument(.nostartfiles) {
       // Locate the Swift registration helper by honouring any explicit
